/*
 *  drivers/s390/net/claw.c
 *    ESCON CLAW network driver
 *
 *    $Revision: 1.17 $
 *
 *  S390 version
 *    Copyright (C) 2002 IBM Corporation
 *    Author(s):    Andy Richter (richtera@us.ibm.com)
 *                  Marc Price (mwprice@us.ibm.com)
 *
 *  Description of the Kernel Parameter
 *
 *   claw=id:0xrrrr:read_buffer:write_buffer:host_name:adapter_name:api_type
 *
 *      eg. claw=0:0x0200:5:5:LINUX390:RS6K:API
 *
 *  where
 *
 *   id                 - device id. if id is 0, device name will be claw0.
 *
 *   rrrr               - the first of 2 consecutive device addresses used
 *                        for the CLAW protocol. The specified address is
 *                        always used as the input channel and the next
 *                        address is used as the output channel.
 *
 *   read_buffer        - specifies number of input buffers to allocate.  
 *   write_buffer       - specifies number of output buffers to allocate. 
 *
 *   host_name          - host name 
 *   adaptor_name       - adaptor name
 *   api_type           - API type TCPIP or API will be sent and expected
 *                        as ws_name
 *
 *   Note the following requirements:
 *   1)  host_name must match the configured host name on the remote
 *       side of the claw device
 *   2)  adaptor_name must match the configured host name on the remote
 *       side of the claw device
 *
 *  Change History
 *    1.00  Initial release shipped
 * 
 */

#include <linux/version.h>


#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/ctype.h>
#include <linux/interrupt.h>
#include <linux/timer.h>
#include <linux/sched.h>
#include <linux/proc_fs.h>
#include <linux/signal.h>
#include <linux/string.h>

#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/ip.h>
#include <linux/if_arp.h>
#include <linux/tcp.h>
#include <linux/skbuff.h>

#include <asm/io.h>
#include <asm/bitops.h>
#include <asm/irq.h>
#include <linux/delay.h>

MODULE_AUTHOR("(C) 2000 IBM Corporation.");
MODULE_DESCRIPTION("Linux for S/390 CLAW Driver");
MODULE_LICENSE("GPL");
MODULE_PARM(claw, "1-"__MODULE_STRING (256)"s");
MODULE_PARM_DESC(claw, 
"Parameters are in the following syntax:\n"
"claw=<device num>:<device addr>:<num read buffers>:"
"<num write buffers>:<host name>:<adaptor name>:<api type>, ...\n"
"E.g.: claw=0:0x1bd0:5:5:LINUX390:RS6K:TCPIP\n");

char *claw[256];


/* Debugging is based on DEBUGMSG, IOTRACE, or FUNCTRACE  options: 
   DEBUGMSG  - Enables output of various debug messages in the code
   IOTRACE   - Enables output of CCW and other IO related traces
   FUNCTRACE - Enables output of function entry/exit trace, and a dump of
               function parameters
   Define any combination of above options to enable tracing
*/

/* following enables tracing */
//#define DEBUGMSG
//#define IOTRACE
//#define FUNCTRACE

#ifdef DEBUGMSG
#define DEBUG
#endif

#ifdef IOTRACE
#define DEBUG
#endif

#ifdef FUNCTRACE
#define DEBUG
#endif

/* CHANDEV Support 
   Comment out the following line to disable CHANDEV support.
   Then the driver will use it's own routines for
   configuration/setup even if chandev is configured in the kernel
*/

#ifdef CONFIG_CHANDEV
#define CLAW_CHANDEV
#endif

#ifdef CLAW_CHANDEV
#include <asm/chandev.h>
#define REQUEST_IRQ chandev_request_irq
#define FREE_IRQ chandev_free_irq
#else
#define REQUEST_IRQ request_irq
#define FREE_IRQ free_irq
#endif

#if LINUX_VERSION_CODE >= 0x020213
#  include <asm/idals.h>
#else
#  define set_normalized_cda(ccw, addr) ((ccw)->cda = (addr),0)
#  define clear_normalized_cda(ccw) ((ccw)->cda = 0)
#endif
#if LINUX_VERSION_CODE < 0x020400
#  define s390_dev_info_t dev_info_t
#  define dev_kfree_skb_irq(a) dev_kfree_skb(a)
#endif

/*******************************************************
*  Define constants                                    *
*                                                      *
********************************************************/
/*-----------------------------------------------------*
*     CCW command codes for CLAW protocol              *
*------------------------------------------------------*/

#define CCW_CLAW_CMD_WRITE           0x01      // Write - not including link
#define CCW_CLAW_CMD_READ            0x02      // read
#define CCW_CLAW_CMD_NOP             0x03      // NOP
#define CCW_CLAW_CMD_SENSE           0x04      // Sense
#define CCW_CLAW_CMD_SIGNAL_SMOD     0x05      // Signal Status Modifier
#define CCW_CLAW_CMD_TIC             0x08      // TIC
#define CCW_CLAW_CMD_READHEADER      0x12      // read header data
#define CCW_CLAW_CMD_READFF          0x22      // read an FF
#define CCW_CLAW_CMD_SENSEID         0xe4      // Sense ID


/*-----------------------------------------------------*
*    CLAW Unique constants                             *
*------------------------------------------------------*/

#define MORE_to_COME_FLAG       0x04     // OR with Write CCW in case of
                                         // more-to-come

#define CLAW_IDLE               0x00     // flag to indicate CLAW is idle
#define CLAW_BUSY               0xff     // flag to indicate CLAW is busy

#define CLAW_PENDING            0x00     // flag to indicate i/o is pending
#define CLAW_COMPLETE           0xff     // flag to indicate i/o completed

/*-----------------------------------------------------*
*     CLAW control comand code                         *
*------------------------------------------------------*/

#define SYSTEM_VALIDATE_REQUEST   0x01  // System Validate request
#define SYSTEM_VALIDATE_RESPONSE  0x02  // System Validate response
#define CONNECTION_REQUEST        0x21  // Connection request
#define CONNECTION_RESPONSE       0x22  // Connection response
#define CONNECTION_CONFIRM        0x23  // Connection confirm
#define DISCONNECT                0x24  // Disconnect
#define CLAW_ERROR                0x41  // CLAW error message
#define CLAW_VERSION_ID           2     // CLAW version ID

/*-----------------------------------------------------*
*  CLAW adater sense bytes                             *
*------------------------------------------------------*/

#define CLAW_ADAPTER_SENSE_BYTE 0x41   // Stop command issued to adapter

/*-----------------------------------------------------*
*      CLAW control command return codes               *
*------------------------------------------------------*/

#define CLAW_RC_NAME_MISMATCH       166  //  names do not match
#define CLAW_RC_WRONG_VERSION       167  //  wrong CLAW version number
#define CLAW_RC_HOST_RCV_TOO_SMALL  180  //  Host maximum receive is less than
                                         //  Linux for zSeries transmit size

/*-----------------------------------------------------*
*      CLAW Constants application name                 *
*------------------------------------------------------*/

#define HOST_APPL_NAME                  "TCPIP   "
#define WS_APPL_NAME_IP_LINK            "TCPIP   "
#define WS_APPL_NAME_API_LINK           "API     "


#define MAX_ENVELOPE_SIZE       65536
#define CLAW_DEFAULT_MTU_SIZE   4096
#define READ                    0
#define WRITE                   1

#define TB_TX                   0          /* sk buffer handling in process */
#define TB_STOP                 1          /* network device stop in process */
#define TB_RETRY                2          /* retry in process */
#define TB_NOBUFFER             3          /* no buffer on free queue */
#define CLAW_MAX_LINK_ID        1
#ifdef CLAW_CHANDEV
#define CLAW_MAX_DEV            256        /* max claw devices */
#else
#define CLAW_MAX_DEV            8
#endif
#define MAX_NAME_LEN            8          /* host name, adapter name length */
#define CLAW_MAX_PARMS_NUM      4          /* numeric parameter */ 
#define CLAW_FRAME_SIZE         4096

/* state machine codes used in claw_irq_handler */

#define CLAW_STOP                0
#define CLAW_START_HALT_IO       1
#define CLAW_START_SENSEID       2
#define CLAW_START_READ          3
#define CLAW_START_WRITE         4

/*-----------------------------------------------------*
*    Lock flag                                         *
*------------------------------------------------------*/
#define LOCK_YES             0
#define LOCK_NO              1


/*******************************************************
*  Define Control Blocks                               *
*                                                      *
********************************************************/
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)

 typedef struct device  net_device;

 typedef struct wait_queue* wait_queue_head_t;
 #define DECLARE_WAITQUEUE(waitqname,waitqtask) \
 	struct wait_queue  waitqname = {waitqtask, NULL }
 #define init_waitqueue_head(x) do { *x = NULL; } while(0)
#else
  typedef struct net_device  net_device;
#endif


/*------------------------------------------------------*/
/*     CLAW header                                      */
/*------------------------------------------------------*/

typedef struct _CLAWH {
        __u16  length;     // length of data read by preceding read CCW
        __u8   opcode;     // equivalent read CCW
        __u8   flag;       // flag of FF to indicate read was completed
} CLAWH, *PCLAWH;

/*------------------------------------------------------*/
/*     CLAW Ending CCWBK                                */
/*------------------------------------------------------*/
typedef struct _ENDCCW {
        __u32     real;            // real address of this block
        __u8      write1;          // write 1 is active
        __u8      read1;           // read 1 is active
        __u16     reserved;        // reserved for future use
        ccw1_t    write1_nop1;     //
        ccw1_t    write1_nop2;     //
        ccw1_t    write2_nop1;     //
        ccw1_t    write2_nop2;     //
        ccw1_t    read1_nop1;      //
        ccw1_t    read1_nop2;      //
        ccw1_t    read2_nop1;      //
        ccw1_t    read2_nop2;      //
} ENDCCW, *PENDCCW;

/*------------------------------------------------------*/
/*     CLAW CCWBK                                       */
/*------------------------------------------------------*/
typedef struct _CCWBK {
        void      *next;        // pointer to next ccw block
        __u32     real;         // real address of this ccw
        void      *p_buffer;    // virtual address of data
        CLAWH     header;       // claw header
        ccw1_t    Write;        // write CCW
        ccw1_t    WreadFF;      // read FF
        ccw1_t    WTIC1;        // TIC
        ccw1_t    read;         // read CCW
        ccw1_t    read_h;        // read header
        ccw1_t    signal;       // signal SMOD
        ccw1_t    RTIC1;        // TIC1
        ccw1_t    r_read_ff;      // read FF
        ccw1_t    RTIC2;        // TIC2
} CCWBK, *PCCWBK;

/*------------------------------------------------------*/
/*     CLAW control block                               */
/*------------------------------------------------------*/
typedef struct _CLAWCTL {
        __u8    command;      // control command
        __u8    version;      // CLAW protocol version
        __u8    linkid;       // link ID
        __u8    correlator;   // correlator
        __u8    rc;           // return code
        __u8    reserved1;    // reserved
        __u8    reserved2;    // reserved
        __u8    reserved3;    // reserved
        __u8    data[24];     // command specific fields
} CLAWCTL, *PCLAWCTL;

/*------------------------------------------------------*/
/*     Data for SYSTEMVALIDATE command                  */
/*------------------------------------------------------*/
typedef struct _SYSVAL  {
        char    WS_name[8];        // Workstation System name 
        char    Host_name[8];      //  Host system name        
        __u16   read_frame_size;   // read frame size
        __u16   Write_frame_size;  // write frame size
        __u8    reserved[4];       // reserved
} SYSVAL, *PSYSVAL;

/*------------------------------------------------------*/
/*     Data for CONNECT command                         */
/*------------------------------------------------------*/
typedef struct _CONNECT  {
        char     WS_name[8];       // Workstation application name 
        char     Host_name[8];     // Host application name      
        __u8     reserved1[4];     // read frame size
        __u8     reserved2[4];     // reserved
} CONNECT, *PCONNECT;

/*------------------------------------------------------*/
/*     Data for CLAW error                              */
/*------------------------------------------------------*/
typedef struct _CLAWERROR  {
        char      reserved1[8];   // reserved
        char      reserved2[8];   // reserved
        char      reserved3[8];   // reserved
} CLAWERROR, *PCLAWERROR;

/*------------------------------------------------------*/
/*     Data buffer for CLAW                             */
/*------------------------------------------------------*/

typedef struct _CLAWBUF  {
        char      buffer[4096];   // data buffer
} CLAWBUF, *PCLAWBUF;

/*------------------------------------------------------*/
/*     Channel control block for read and write channel */
/*------------------------------------------------------*/

typedef struct _CHBK {
        unsigned int        devno;
        int                 irq;
        __u32               IO_active;
        __u8                claw_state;
        devstat_t           *devstat;
        net_device          *dev;  /* backward pointer to the network device */
        wait_queue_head_t   wait;
        struct tq_struct    tq;
        struct timer_list   timer;
        unsigned long       flag_a;    /* atomic flags */
#define CLAW_BH_ACTIVE      0
        unsigned long       flag_b;    /* atomic flags */
#define CLAW_WRITE_ACTIVE   0
        __u8                last_dstat;
        __u8                flag;
#define CLAW_WRITE      0x01      /* - Set if this is a write channel */
#define CLAW_TIMER      0x80      /* - Set if timer made the wake_up  */
} CHBK, *PCHBK;

/*--------------------------------------------------------------*
*           CLAW  environment block                             *
*---------------------------------------------------------------*/

typedef struct _CLAW_ENV {
        char                    devname[9];     // device name    
        unsigned int            devno[2];       // device number
        char                    host_name[9];    // Host name
        char                    adapter_name[9]; // adapter name
        char                    ApiType[9];     // API type either TCPIP or API
        void                    *p_priv;        // privptr 
        __u16                   readBuffers;    // read buffer number
        __u16                   WriteBuffers;   // write buffer number
        __u16                   readSize;       // read buffer size
        __u16                   write_size;      // write buffer size
        __u16                   dev_id;         // device ident
        __u8                    in_use;         // device active flag
        net_device              *dev;           // backward pointer to the
	                                        // network device 
} CLAW_ENV, *PENV;

/*--------------------------------------------------------------*
*           CLAW  main control block                            *
*---------------------------------------------------------------*/

typedef struct _CLAW_PRIVBK {
        void       *p_buffer_ccw;
        __u32      p_buffer_ccw_num;
        void       *p_buffer_read;
        __u32      p_buffer_read_num;
        __u32      p_buffer_pages_perread;  
        void       *p_buffer_write;
        __u32      p_buffer_write_num;
        __u32      p_buffer_pages_perwrite;
        long       ActiveLinkID;           // Active logical link ID
        PCCWBK     p_write_free_chain;     // pointer to free ccw chain
        PCCWBK     p_write_active_first;   // pointer to the first write ccw
        PCCWBK     p_write_active_last;    // pointer to the last write ccw
        PCCWBK     p_read_active_first;    // pointer to the first read ccw
        PCCWBK     p_read_active_last;     // pointer to the last read ccw
        PENDCCW    p_end_ccw;              // pointer to ending ccw
        PCCWBK     pClawSignalBlock;       // pointer to signal block
        __u32      write_free_count;       // number of free bufs on write
	                                   // free chain
        struct     net_device_stats  stats;  // device status
        CHBK       channel[2];             // Channel control block
        __u8       MoreToComeSkipping;
        int        more_to_come_offset;
        int        more_to_come_logical_link;
        void       *p_more_to_come_envelope;
        CLAWCTL    ctl_bk;
        PENV       p_env;
        __u8       SYSTEM_VALIDATE_COMP;
        __u8       RELEASE_PEND;
#ifdef CLAW_CHANDEV
        __u8      checksum_received_ip_pkts;
#endif
        ENDCCW     EndCCW;
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,3,0)
        __u32      tbusy;
#endif
} CLAW_PRIVBK, *PPRIV;


/*-----------------------------------------------------*
*      /proc filesystem structure                      *
*------------------------------------------------------*/

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
int claw_read_proc(char *, char **, off_t, int, int);
#else
int claw_read_proc(char *, char **, off_t, int);
#endif
struct proc_dir_entry *claw_proc_entry = NULL;




/************************************************************/
/* define global constants                                  */
/************************************************************/

#define CCWBK_SIZE sizeof(CCWBK)

/***********************************************************/
/*    Define control blocks                                */
/***********************************************************/

int num_claw_devices;
static CLAW_ENV     Env[CLAW_MAX_DEV];
 
static __inline__ void claw_set_busy(net_device *dev)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
 dev->tbusy=1;
#else
 ((CLAW_PRIVBK *)dev->priv)->tbusy=1;
#endif
 eieio();
}

static __inline__ void claw_clear_busy(net_device *dev)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
 dev->tbusy=0;
#else
 ((CLAW_PRIVBK *)dev->priv)->tbusy=0;
#endif
 eieio();
}

static __inline__ int claw_check_busy(net_device *dev)
{
 eieio();
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
 return(dev->tbusy);
#else 
 return(((CLAW_PRIVBK *)dev->priv)->tbusy);
#endif
}

static __inline__ void claw_setbit_busy(int nr,net_device *dev)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
 set_bit(nr,(void *)&dev->tbusy);
#else
 set_bit(nr, (void *)&(((CLAW_PRIVBK *)dev->priv)->tbusy));
#endif
}

static __inline__ void claw_clearbit_busy(int nr,net_device *dev)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
 clear_bit(nr,(void *)&dev->tbusy);
#else
 clear_bit(nr,(void *)&(((CLAW_PRIVBK *)dev->priv)->tbusy));
#endif
}

static __inline__ int claw_test_and_setbit_busy(int nr,net_device *dev)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
 return(test_and_set_bit(nr,(void *)&dev->tbusy));
#else
 return(test_and_set_bit(nr, (void *)&(((CLAW_PRIVBK *)dev->priv)->tbusy)));
#endif
 return 0;
}


/* Functions for the DEV methods */

int claw_probe(net_device *dev);
static int claw_tx(struct sk_buff *skb, net_device *dev);
static int claw_change_mtu( net_device *dev, int new_mtu);
static int claw_open(net_device *dev);
static void claw_irq_handler (int irq, void *initparm, struct pt_regs *regs);
static void claw_irq_bh ( PCHBK p_ch);
static int claw_release(net_device *dev);
static void claw_write_retry ( PCHBK p_ch );
static void claw_write_next ( PCHBK p_ch );
static void claw_timer ( PCHBK p_ch );

/* Functions */
static int add_claw_reads(net_device *dev, PCCWBK p_first, PCCWBK p_last);
static void inline ccw_check_return_code (net_device *dev, int return_code);
static void inline ccw_check_unit_check (net_device *dev,  devstat_t *devstat);
static int inline extract_dev_id(char *name);
static int find_link(net_device *dev, char *host_name, char *ws_name );
static int claw_hw_tx(struct sk_buff *skb, net_device *dev, long linkid);
static int init_ccw_bk(net_device *dev);
static void probe_error( net_device *dev);
struct net_device_stats *claw_stats(net_device *dev);
#ifdef DEBUG
static void dumpit (char *buf, int len);
#endif

/*   Functions for System Validate  */
static int process_claw_control( net_device *dev, PCCWBK p_ccw);
static int send_claw_control(net_device *dev, __u8 type, __u8 link,
			     __u8 correlator, __u8 rc , char *local_name, 
			     char *remote_name);
static int send_connect_request(net_device *dev, __u8 link);
static int send_disconnect(net_device *dev, PCLAWCTL p_ctl);
static int send_system_validate_response(net_device *dev, PCLAWCTL p_ctl,
					 __u32 return_code);
static int start_connect_request(net_device *dev );
static void start_read ( net_device *dev, int lock );
static void start_output_io( net_device *dev );
static void free_write_buffers( net_device *dev );

/* Functions for unpack reads   */
static void unpack_read (net_device *dev );


#ifndef CLAW_CHANDEV
/************************************************************/
/* The devices                                              */
/************************************************************/
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0))
static char claw_names[CLAW_MAX_DEV][8];
#endif
net_device claw_devs[CLAW_MAX_DEV] = 
{
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0))
        {
                name: &claw_names[0][0],  /* Name filled in at load time  */
                init: claw_probe          /* probe function               */
        },
        {
                name: &claw_names[1][0],  /* Name filled in at load time  */
                init: claw_probe          /* probe function               */
        },
        {
                name: &claw_names[2][0],  /* Name filled in at load time  */
                init: claw_probe          /* probe function               */
        },
        {
                name: &claw_names[3][0],  /* Name filled in at load time  */
                init: claw_probe          /* probe function               */
        },
        {
                name: &claw_names[4][0],  /* Name filled in at load time  */
                init: claw_probe          /* probe function               */
        },
        {
                name: &claw_names[5][0],  /* Name filled in at load time  */
                init: claw_probe          /* probe function               */
        },
        {
                name: &claw_names[6][0],  /* Name filled in at load time  */
                init: claw_probe          /* probe function               */
        },
        {
                name: &claw_names[7][0],  /* Name filled in at load time  */
                init: claw_probe          /* probe function               */
        },
#else /* (LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)) */
        {
                name: &claw_names[0][0],  /* Name filled in at load time  */
                init: claw_probe          /* probe function               */
        },
        {
                name: &claw_names[1][0],  /* Name filled in at load time  */
                init: claw_probe          /* probe function               */
        },
        {
                name: &claw_names[2][0],  /* Name filled in at load time  */
                init: claw_probe          /* probe function               */
        },
        {
                name: &claw_names[3][0],  /* Name filled in at load time  */
                init: claw_probe          /* probe function               */
        },
        {
                name: &claw_names[4][0],  /* Name filled in at load time  */
                init: claw_probe          /* probe function               */
        },
        {
                name: &claw_names[5][0],  /* Name filled in at load time  */
                init: claw_probe          /* probe function               */
        },
        {
                name: &claw_names[6][0],  /* Name filled in at load time  */
                init: claw_probe          /* probe function               */
        },
        {
                name: &claw_names[7][0],  /* Name filled in at load time  */
                init: claw_probe          /* probe function               */
        },
#endif
};
#endif

/*
*
*       Key functions
*/


#ifdef CLAW_CHANDEV

/*----------------------------------------------------------------*
 *   claw_chandev_probe                                           *
 *      this function is called for each claw network device,     *
 *      which is defined in the /etc/chandev.conf                 *
 *      This probe routine is called for CHANDEV configuration    *
 *----------------------------------------------------------------*/
int claw_chandev_probe(chandev_probeinfo *info)                 
{       
        int             rc;
        int             dev_id;
        PPRIV           privptr;
        PENV            p_env;
        int             c,numbuffers;
        net_device      *dev;

#ifdef DEBUGMSG
        printk(KERN_INFO "claw: claw_chandev_probe() > enter  \n");
        printk(KERN_INFO "claw: variable info =\n");
        dumpit((char *)info, sizeof(chandev_probeinfo));
#endif

        dev = kmalloc(sizeof(net_device), GFP_KERNEL);
        if (!dev)
                return -ENOMEM;
        memset(dev, 0, sizeof(net_device));
        
        if (!chandev_build_device_name(info, dev->name, "claw", 1))
           return -ENODEV;
                        
        
        dev_id=num_claw_devices;                // first device is 0 and so on 
        ++num_claw_devices;
        
        /* fill up env_bk    */
        for (c = 0; c <= CLAW_MAX_DEV; c++) {
                p_env = Env + c;
                if (!(p_env->in_use)) {
                        p_env->in_use = 1;
                        break;
                }
        }
        p_env->devno[READ]  = info->read.devno;
        p_env->devno[WRITE] = info->write.devno;
        p_env->readSize     = CLAW_FRAME_SIZE;
        p_env->write_size    = CLAW_FRAME_SIZE;

        /*
        * get names
        */
        
        memset(p_env->host_name, 0x20, MAX_NAME_LEN);
        memset(p_env->adapter_name, 0x20, MAX_NAME_LEN);
        memset(p_env->ApiType, 0x20,  MAX_NAME_LEN);
        memcpy(p_env->host_name, info->claw.host_name,
	       strlen(info->claw.host_name));
        memcpy(p_env->adapter_name, info->claw.adapter_name,
	       strlen(info->claw.adapter_name));
        memcpy(p_env->ApiType, info->claw.api_type,
	       strlen(info->claw.api_type));
                
        /*
        *  calculate buffers   
        */
                
        numbuffers = 20;            // default to memory_usage_in_k as 160
        if (info->memory_usage_in_k >= 8) 
                numbuffers = (int) (info->memory_usage_in_k /
				    (int)(CLAW_FRAME_SIZE/1024*2));
        p_env->readBuffers = numbuffers;
        p_env->WriteBuffers = numbuffers;
        info->memory_usage_in_k=-((CLAW_FRAME_SIZE/1024)*
				  (p_env->readBuffers+p_env->WriteBuffers));
        if ( (p_env->devno[READ] == -ENODEV) ||
	     (p_env->devno[WRITE] == -ENODEV) ) {
#ifdef DEBUGMSG
                printk(KERN_INFO "claw: bad read/write device number\n");
#endif
#ifdef FUNCTRACE
                printk(KERN_INFO "claw: chandev_claw_probe() > exit on "
		       "line %d, rc =  ENODEV\n", __LINE__);
#endif
                return -ENODEV;
        }
#ifdef DEBUGMSG 
        printk(KERN_INFO "%s: Buffers > read: %d  Write: %d  mem_in_k: %d \n",
                        dev->name,p_env->readBuffers,p_env->WriteBuffers,
			info->memory_usage_in_k);
#endif
        dev->priv = kmalloc(sizeof(CLAW_PRIVBK), GFP_KERNEL);
        if (dev->priv == NULL) {
#ifdef DEBUGMSG
                printk(KERN_INFO "%s: Unable to allocate kernel memory\n",
		       dev->name);
#endif
#ifdef FUNCTRACE
                printk(KERN_INFO "%s: chandev_claw_probe() > exit on line "
		       "%d, rc = ENOMEM\n",dev->name,__LINE__);
#endif
                return -ENOMEM;
        }
        memset(dev->priv, 0x00, sizeof(CLAW_PRIVBK));  
        privptr = (PPRIV) (dev->priv);

        privptr->channel[READ].devno = p_env->devno[READ];
        privptr->channel[READ].irq = info->read.irq;
        privptr->channel[WRITE].devno = p_env->devno[WRITE];
        privptr->channel[WRITE].irq = info->write.irq;

        privptr->p_env=p_env;
        privptr->checksum_received_ip_pkts=info->checksum_received_ip_pkts;
        p_env->p_priv=privptr;
        printk(KERN_INFO "%s: CLAW read dev: %04x irq: %04x - write dev: "
	       "%04x irq: %04x \n",
                dev->name, privptr->channel[READ].devno,
		privptr->channel[READ].irq,
                privptr->channel[WRITE].devno, privptr->channel[WRITE].irq); 
        dev->mtu             = CLAW_DEFAULT_MTU_SIZE;
        dev->hard_start_xmit = claw_tx;
        dev->open            = claw_open;
        dev->stop            = claw_release;
        dev->get_stats       = claw_stats;
        dev->change_mtu      = claw_change_mtu;
        dev->hard_header_len = 0;
        dev->addr_len        = 0;
        dev->type            = ARPHRD_SLIP;
        dev->tx_queue_len    = 100;
        dev_init_buffers(dev);
        dev->flags           = IFF_POINTOPOINT | IFF_NOARP;   
        privptr->SYSTEM_VALIDATE_COMP=0;
        privptr->RELEASE_PEND=0;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
        dev->start=1;
#endif
        p_env->dev = dev;
                
        for (c = READ; c <= WRITE; c++) {
                privptr->channel[c].devstat = kmalloc(sizeof(devstat_t),
						      GFP_KERNEL);
                if (privptr->channel[c].devstat == NULL){
                        if (c == WRITE) { 
                                probe_error(dev);
#ifdef DEBUGMSG
                                printk(KERN_INFO "%s: Unable to allocate "
				       "kernel memory\n",dev->name);
#endif
#ifdef FUNCTRACE
                                printk(KERN_INFO "%s: chandev_claw_probe() "
				       "> exit on line %d, rc = ENOMEM\n",
				       dev->name,__LINE__);
#endif
                                return -ENOMEM;
                        }
                }
                memset(privptr->channel[c].devstat, 0x00, sizeof(devstat_t));

                rc = REQUEST_IRQ(privptr->channel[c].irq,
				 (void *)claw_irq_handler, 
                                SA_INTERRUPT, dev->name,
				privptr->channel[c].devstat);

                if (rc) { 
                        printk(KERN_WARNING "%s: request_irq failed rc=%02x\n",
			       dev->name, -rc);
                        probe_error(dev);
#ifdef FUNCTRACE
                        printk(KERN_INFO "%s: chandev_claw_probe() > exit "
			       "on line %d, rc = EBUSY\n",dev->name ,__LINE__);
#endif
                        return -EBUSY;
                }
        } /* end of for (c = 0; c < 2; c++) */

        privptr->p_more_to_come_envelope= kmalloc( MAX_ENVELOPE_SIZE, 
						   GFP_KERNEL);

        if (privptr->p_more_to_come_envelope==NULL) {
                probe_error(dev);
#ifdef FUNCTRACE
                printk(KERN_INFO "%s: chandev_claw_probe() > exit on line %d, "
		       "rc = ENOMEM\n",dev->name ,__LINE__);              
#endif
                return -ENOMEM;
        }

        if ( (rc = register_netdev(dev)) != 0 ) 
        {
                printk(KERN_WARNING "%s: chandev_claw_probe() error %i "
		       "registering device \"%s\"\n",
                        dev->name,rc, dev->name);
                probe_error(dev);
                return -ENODEV;
        }
        
        chandev_initdevice(info, dev, 0, dev->name,
                           chandev_category_network_device,
                           (chandev_unregfunc)unregister_netdev);
        
                
#ifdef FUNCTRACE
        printk(KERN_INFO "%s: chandev_claw_probe() > exit on line %d, "
	       "rc = 0\n",dev->name,__LINE__);
#endif

        return 0;
}  /*  end of chandev_claw_probe       */

#else

/*----------------------------------------------------------------*
 *   claw_probe                                                   *
 *      this function is called for each claw network device,     *
 *      which is defined in the /init/main.c                      *
 *      This probe routine is called when CHANDEV configuration   *
 *      is NOT enabled                                            * 
 *----------------------------------------------------------------*/
int claw_probe(net_device *dev)
{       
        int             rc;
        int             dev_id;
        PPRIV           privptr;
        PENV            p_env;
        int             i,c;

#ifdef DEBUGMSG
        printk(KERN_INFO "%s: claw_probe() > enter  \n",dev->name);
        printk(KERN_INFO "%s: variable dev =\n",dev->name);
        dumpit((char *)dev, sizeof(net_device));
#endif
        p_env=(PENV)&Env;
        dev_id=extract_dev_id(dev->name);
        for (i=0; i< dev_id ; i++) {
                ++p_env;
        }

        if ( (p_env->devno[READ] == -ENODEV) ||
	     (p_env->devno[WRITE] == -ENODEV) ) {
#ifdef DEBUGMSG
                printk(KERN_INFO "%s: bad read/write device number\n",
		       dev->name);
#endif
#ifdef FUNCTRACE
                printk(KERN_INFO "%s: claw_probe() > exit on line %d, "
		       "rc =  ENODEV\n",dev->name,__LINE__);
#endif
                return -ENODEV;
        }
        dev->priv = kmalloc(sizeof(CLAW_PRIVBK), GFP_KERNEL);
        if (dev->priv == NULL) {
#ifdef DEBUGMSG
                printk(KERN_INFO "%s: Unable to allocate kernel memory\n",
		       dev->name);
#endif
#ifdef FUNCTRACE
                printk(KERN_INFO "%s: claw_probe() > exit on line %d, "
		       "rc = ENOMEM\n",dev->name,__LINE__);
#endif
                return -ENOMEM;
        }
        memset(dev->priv, 0x00, sizeof(CLAW_PRIVBK));  
        privptr = (PPRIV) (dev->priv);
        for (c = READ; c <= WRITE; c++) {
                privptr->channel[c].devstat = kmalloc(sizeof(devstat_t),
						      GFP_KERNEL);
                if (privptr->channel[c].devstat == NULL){
                        if (c == WRITE) { 
                                probe_error(dev);
#ifdef DEBUGMSG
                                printk(KERN_INFO "%s: Unable to allocate "
				       "kernel memory\n",dev->name);
#endif
#ifdef FUNCTRACE
                                printk(KERN_INFO "%s: claw_probe() > exit "
				       "on line %d, rc = ENOMEM\n",
				       dev->name,__LINE__);
#endif
                                return -ENOMEM;
                        }
                }
                memset(privptr->channel[c].devstat, 0x00, sizeof(devstat_t));

                rc = REQUEST_IRQ(get_irq_by_devno(p_env->devno[c]),
				 (void *)claw_irq_handler, 
				 SA_INTERRUPT, dev->name,
				 privptr->channel[c].devstat);

                if (rc) { 
                        printk(KERN_WARNING "%s: request_irq failed rc=%02x\n",
			       dev->name, -rc);
                        probe_error(dev);
#ifdef FUNCTRACE
                        printk(KERN_INFO "%s: claw_probe() > exit on "
			       "line %d, rc = EBUSY\n",dev->name,__LINE__);
#endif
                        return -EBUSY;
                }
        } /* end of for (c = 0; c <= 1; c++) */

        privptr->p_more_to_come_envelope= kmalloc( MAX_ENVELOPE_SIZE, 
						   GFP_KERNEL);

        if (privptr->p_more_to_come_envelope==NULL) {
                probe_error(dev);
#ifdef FUNCTRACE
                printk(KERN_INFO "%s: claw_probe() > exit on line %d, "
		       "rc = ENOMEM\n",dev->name,__LINE__);              
#endif
                return -ENOMEM;
        }


        privptr->channel[READ].devno = p_env->devno[READ];

        privptr->channel[READ].irq = get_irq_by_devno(p_env->devno[READ]);
        privptr->channel[WRITE].devno = p_env->devno[WRITE];
        privptr->channel[WRITE].irq = get_irq_by_devno(p_env->devno[WRITE]);

        privptr->p_env=p_env;
        p_env->p_priv=privptr;
        p_env->in_use = 1;
        printk(KERN_INFO "%s: CLAW read dev: %04x irq: %04x - write dev: "
	       "%04x irq: %04x \n",
                dev->name, privptr->channel[READ].devno,
		privptr->channel[READ].irq,
                privptr->channel[WRITE].devno, privptr->channel[WRITE].irq); 
        dev->mtu             = CLAW_DEFAULT_MTU_SIZE;
        dev->hard_start_xmit = claw_tx;
        dev->open            = claw_open;
        dev->stop            = claw_release;
        dev->get_stats       = claw_stats;
        dev->change_mtu      = claw_change_mtu;
        dev->hard_header_len = 0;
        dev->addr_len        = 0;
        dev->type            = ARPHRD_SLIP;
        dev->tx_queue_len    = 100;
        dev_init_buffers(dev);
        dev->flags           = IFF_POINTOPOINT | IFF_NOARP;   
        privptr->SYSTEM_VALIDATE_COMP=0;
        privptr->RELEASE_PEND=0;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
        dev->start=1;
#endif
#ifdef FUNCTRACE
        printk(KERN_INFO "%s: claw_probe() > exit on line %d, rc = 0\n",
	       dev->name,__LINE__);
#endif

        return 0;
}  /*  end of claw_probe       */
#endif

/*-------------------------------------------------------------------*
 *   claw_tx                                                         *
 *-------------------------------------------------------------------*/

static int claw_tx(struct sk_buff *skb, net_device *dev)
{
        int             rc;
        PPRIV           privptr=dev->priv;
        unsigned long   saveflags;
        PCHBK       p_ch;       

#ifdef FUNCTRACE
        printk(KERN_INFO "%s: claw_tx() > enter\n",dev->name);
#endif
        p_ch=&privptr->channel[WRITE];


        if (skb == NULL) { 
                printk(KERN_WARNING "%s: null pointer passed as sk_buffer\n",
		       dev->name);
                privptr->stats.tx_dropped++;
#ifdef FUNCTRACE
                printk(KERN_INFO "%s: claw_tx() > exit on line %d, "
		       "rc = EIO\n",dev->name, __LINE__);
#endif
                return -EIO;
        }
#ifdef FUNCTRACE
        printk(KERN_INFO "%s: variable sk_buff=\n",dev->name);
        dumpit((char *) skb, sizeof(struct sk_buff));
        printk(KERN_INFO "%s: variable dev=\n",dev->name);
        dumpit((char *) dev, sizeof(net_device));
#endif

        s390irq_spin_lock_irqsave(p_ch->irq, saveflags);
        rc=claw_hw_tx( skb, dev, 1 );
        s390irq_spin_unlock_irqrestore(p_ch->irq, saveflags);

#ifdef FUNCTRACE
        printk(KERN_INFO "%s: claw_tx() > exit on line %d, rc = %d\n",
	       dev->name, __LINE__, rc);
#endif
        return rc;
}   /*  end of claw_tx */



/*-------------------------------------------------------------------*
 *   claw_change_mtu                                                 *
 *                                                                   *
 *-------------------------------------------------------------------*/

static int claw_change_mtu( net_device *dev, int new_mtu)
{
#ifdef FUNCTRACE
        printk(KERN_INFO "%s: claw_change_mtu() > enter  \n",dev->name);
        printk(KERN_INFO "variable dev =\n");
        dumpit((char *) dev, sizeof(net_device));
        printk(KERN_INFO "variable new_mtu = %d\n", new_mtu);
#endif
        if ((new_mtu < 60) || (new_mtu > 65528)) {
#ifdef FUNCTRACE
                printk(KERN_INFO "claw: claw_change_mtu() > exit on line %d, "
		       "rc=EINVAL\n", __LINE__);
#endif
                return -EINVAL;
        }
        dev->mtu = new_mtu;
#ifdef FUNCTRACE
        printk(KERN_INFO "%s: claw_change_mtu() > exit on line %d\n",
	       dev->name, __LINE__);
#endif
        return 0;
}  /*   end of claw_change_mtu */


/*-------------------------------------------------------------------*
 *   claw_open                                                       *
 *                                                                   *
 *-------------------------------------------------------------------*/
static int claw_open(net_device *dev)
{

        int     rc;
        int     i;
        __u8    flags = 0x00;
        unsigned long   saveflags;
        unsigned long   parm;
        PPRIV   privptr=dev->priv;
        DECLARE_WAITQUEUE(wait, current);
        struct timer_list  timer;
        PCCWBK  p_buf;

#ifdef FUNCTRACE
        printk(KERN_INFO "claw: claw_open() > enter  \n");
        printk(KERN_INFO "dev\n");
        dumpit((char *) dev, sizeof(net_device));
#endif

        /*   allocate and initialize CCW blocks */
        rc=init_ccw_bk(dev);    
        if (rc) {
#ifdef FUNCTRACE
                printk(KERN_INFO "%s: claw_open() > exit on line %d, "
		       "rc=ENOMEM\n",dev->name,__LINE__);
#endif
                return -ENOMEM;
        }

        privptr->SYSTEM_VALIDATE_COMP=0;
        privptr->RELEASE_PEND=0;

        claw_set_busy(dev);
        privptr = (PPRIV) (dev->priv);    
        privptr->channel[READ].flag  = 0x00;
        privptr->channel[WRITE].flag = CLAW_WRITE;

        for ( i = 0; i < 2;  i++) { 
                init_waitqueue_head(&privptr->channel[i].wait);	
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
                privptr->channel[i].tq.next = NULL;
#endif
                privptr->channel[i].tq.sync = 0;
                privptr->channel[i].tq.routine = (void *)(void *)claw_irq_bh;
                privptr->channel[i].tq.data = &privptr->channel[i]; 
                privptr->channel[i].dev = dev;
                privptr->channel[i].flag_a = 0;
                privptr->channel[i].IO_active = 0;
                privptr->channel[i].flag  &= ~CLAW_TIMER;
                init_timer(&timer);
                timer.function = (void *)claw_timer; 
                timer.data = (unsigned long)(&privptr->channel[i]);
                timer.expires = jiffies + 15*HZ;      /* time to connect with
							 the remote side */
                add_timer(&timer);        
                s390irq_spin_lock_irqsave(privptr->channel[i].irq, saveflags);
                parm = (unsigned long) &privptr->channel[i];
                privptr->channel[i].claw_state = CLAW_START_HALT_IO;
                rc = halt_IO(privptr->channel[i].irq, parm, flags);       
                add_wait_queue(&privptr->channel[i].wait, &wait);
                current->state = TASK_INTERRUPTIBLE;
                s390irq_spin_unlock_irqrestore(privptr->channel[i].irq,
					       saveflags);        
                schedule(); 
                remove_wait_queue(&privptr->channel[i].wait, &wait);         
                if(rc != 0)
                        ccw_check_return_code(dev, rc);
                if((privptr->channel[i].flag & CLAW_TIMER) == 0x00)
                        del_timer(&timer);
        }

        if ((((privptr->channel[READ].last_dstat |
	       privptr->channel[WRITE].last_dstat) & 
	      ~(DEV_STAT_CHN_END | DEV_STAT_DEV_END)) != 0x00) ||
	    (((privptr->channel[READ].flag | privptr->channel[WRITE].flag) &
	      CLAW_TIMER) != 0x00)) {
#ifdef DEBUGMSG
                printk(KERN_WARNING "%s: channel problems during open - "
		       "read: %02x -  write: %02x\n",
                        dev->name, privptr->channel[READ].last_dstat,
			privptr->channel[WRITE].last_dstat);
#endif 
                printk(KERN_INFO "%s: remote side is not ready\n", dev->name);
            
                for ( i = 0; i < 2;  i++) {
                        s390irq_spin_lock_irqsave(privptr->channel[i].irq,
						  saveflags);
                        parm = (unsigned long) &privptr->channel[i];
                        privptr->channel[i].claw_state = CLAW_STOP;
                        rc = halt_IO(privptr->channel[i].irq, parm, flags);
                        s390irq_spin_unlock_irqrestore(privptr->channel[i].irq,
						       saveflags);
                        if (rc != 0) {
                                ccw_check_return_code(dev, rc);
                        }
                }
                free_pages((unsigned long)privptr->p_buffer_ccw,
			   (int)privptr->p_buffer_ccw_num);
                if (privptr->p_env->readSize < PAGE_SIZE) {
                        free_pages((unsigned long)privptr->p_buffer_read,
				   (int)privptr->p_buffer_read_num);
                }
                else { 
                        p_buf=privptr->p_read_active_first;
                        while (p_buf!=NULL) {
                                free_pages((unsigned long)p_buf->p_buffer,
					   (int)privptr->
					   p_buffer_pages_perread);
                                p_buf=p_buf->next;
                        }
                }
                if (privptr->p_env->write_size < PAGE_SIZE ) {
                        free_pages((unsigned long)privptr->p_buffer_write,
				   (int)privptr->p_buffer_write_num);
                }
                else {
                        p_buf=privptr->p_write_active_first;
                        while (p_buf!=NULL) {
                                free_pages((unsigned long)p_buf->p_buffer,
					   (int)privptr->
					   p_buffer_pages_perwrite);
                                p_buf=p_buf->next;
                        }
                }
                privptr->p_buffer_ccw=NULL;
                privptr->p_buffer_read=NULL;
                privptr->p_buffer_write=NULL;
                claw_clear_busy(dev);
#ifdef FUNCTRACE
                printk(KERN_INFO "%s: claw_open() > exit on line %d, "
		       "rc=EIO\n",dev->name,__LINE__);
#endif
                return -EIO;
        }

        /*   Send SystemValidate command */
        
        claw_clear_busy(dev);
        
#ifdef FUNCTRACE
        printk(KERN_INFO "%s: claw_open() > exit on line %d\n",
	       dev->name,__LINE__);
#endif
        MOD_INC_USE_COUNT;
        return 0;
}    /*     end of claw_open    */

/*-------------------------------------------------------------------*
*                                                                    *
*       claw_irq_handler                                             *
*                                                                    *
*--------------------------------------------------------------------*/
static void claw_irq_handler (int irq, void *initparm, struct pt_regs *regs)
{
        PCHBK                   p_ch = NULL;
        PPRIV                   privptr = NULL;
        net_device              *dev = NULL;            
        devstat_t *devstat =    ((devstat_t *)initparm);
        PENV                    p_env;
        PCHBK                   p_ch_r=NULL;


#ifdef FUNCTRACE
        printk(KERN_INFO "claw: claw_irq_handler() > enter  \n");
        printk(KERN_INFO "variable irq = %d, initparm =\n", irq);
        dumpit((char *) initparm, sizeof(devstat_t));
#endif

        /* Bypass all 'unsolicited interrupts' */
        if (devstat->intparm == 0) {
#ifdef DEBUGMSG
                printk(KERN_INFO "claw: unsolicited interrupt for device: "
		       "%04x received c-%02x d-%02x f-%08x\n",
                        devstat->devno, devstat->cstat,
			devstat->dstat, devstat->flag);
#endif 
#ifdef FUNCTRACE
                printk(KERN_INFO "claw: claw_irq_handler() > exit on "
		       "line %d\n", __LINE__);
#endif
                return;
        }

        p_ch = (PCHBK) (devstat->intparm);
        dev = (net_device *) p_ch->dev;
        privptr = dev->priv;
        p_env=privptr->p_env;

#ifdef DEBUGMSG
        printk(KERN_INFO "%s: interrupt for device: %04x received "
	       "c-%02x d-%02x f-%08x state-%02x\n",
                dev->name, p_ch->devno, devstat->cstat,
		devstat->dstat, devstat->flag, p_ch->claw_state);
#endif 

        /* Check for good subchannel return code, otherwise error message */
        if (devstat->cstat  &&  !(devstat->cstat & SCHN_STAT_PCI)) {
                printk(KERN_INFO "%s: subchannel check for device: "
		       "%04x - %02x %02x %08x\n", 
                        dev->name, p_ch->devno, devstat->cstat,
			devstat->dstat, devstat->flag);
#ifdef FUNCTRACE
                printk(KERN_INFO "%s: claw_irq_handler() > exit on "
		       "line %d\n",dev->name,__LINE__);
#endif                  
                return;
        }
        
        /* Check the reason-code of a unit check */
        if (devstat->dstat & DEV_STAT_UNIT_CHECK) {
                ccw_check_unit_check(dev, devstat);
        }

        /* State machine to bring the connection up / down and to restart */ 
        p_ch->last_dstat = devstat->dstat;

        switch (p_ch->claw_state) {
                case CLAW_STOP:       /* HALT_IO issued by claw_release
					 (halt sequence) */
#ifdef DEBUGMSG
                        printk(KERN_INFO "%s: CLAW_STOP enter\n", dev->name);
#endif
                        if (!devstat->flag & DEVSTAT_FINAL_STATUS) {
#ifdef FUNCTRACE
                                printk(KERN_INFO "%s: claw_irq_handler() "
				       "> exit on line %d\n",
				       dev->name,__LINE__);
#endif
                                return;
                        }
                        wake_up(&p_ch->wait);   /* wake up claw_release */
                                
#ifdef DEBUGMSG
                        printk(KERN_INFO "%s: CLAW_STOP exit\n", dev->name);
#endif          
#ifdef FUNCTRACE
                        printk(KERN_INFO "%s: claw_irq_handler() "
			       "> exit on line %d\n",dev->name,__LINE__);
#endif
                        return;
                        
                case CLAW_START_HALT_IO:  /* HALT_IO issued by claw_open
					     (start sequence) */
#ifdef DEBUGMSG
                        printk(KERN_INFO "%s: process CLAW_STAT_HALT_IO\n",
			       dev->name);
#endif 
                        if (!devstat->flag & DEVSTAT_FINAL_STATUS) {
#ifdef FUNCTRACE
                                printk(KERN_INFO "%s: claw_irq_handler() "
				       "> exit on line %d\n",
				       dev->name,__LINE__);
#endif
                                return;
                        }
                        if (!p_ch->flag & CLAW_WRITE) {
                                p_ch->claw_state = CLAW_START_READ;
                                wake_up(&p_ch->wait);  /* wake up claw_open
							  (READ) */    
                        } 
                        else {
                                p_ch->claw_state = CLAW_START_WRITE;
                                /*      send SYSTEM_VALIDATE          */
                                start_read(dev, LOCK_NO);
                                send_claw_control(dev,SYSTEM_VALIDATE_REQUEST,
						  0, 0, 0, p_env->host_name,
						  p_env->adapter_name);
                        }
#ifdef DEBUGMSG
                        printk(KERN_INFO "%s: process CLAW_STAT_HALT_IO "
			       "exit\n", dev->name);
#endif 
#ifdef FUNCTRACE
                        printk(KERN_INFO "%s: claw_irq_handler() "
			       "> exit on line %d\n",dev->name,__LINE__);
#endif
                        return;
                case CLAW_START_READ:
                        if (devstat->dstat & DEV_STAT_UNIT_CHECK) {
                                clear_bit(0, (void *)&p_ch->IO_active);
                                if ((devstat->ii.sense.data[0]&0x41) == 0x41 ||
                                    (devstat->ii.sense.data[0]&0x40) == 0x40 ||
                                    (devstat->ii.sense.data[0])      == 0)
                                {
                                        privptr->stats.rx_errors++;
                                        printk(KERN_INFO "%s: Restart is "
					       "required after remote side "
					       "recovers \n",dev->name);
                                }
#ifdef FUNCTRACE
                                printk(KERN_INFO "%s: claw_irq_handler() "
				       "> exit on line %d\n",
				       dev->name,__LINE__);
#endif
                                        return;
                        }
                        if ((devstat->cstat & SCHN_STAT_PCI) &&
			    (devstat->dstat==0)) {
                                if (test_and_set_bit(CLAW_BH_ACTIVE,
						     (void *)&p_ch->flag_a)
				    == 0) {
                                        queue_task(&p_ch->tq, &tq_immediate);
                                        mark_bh(IMMEDIATE_BH);
                                }
#ifdef FUNCTRACE
                                printk(KERN_INFO "%s: claw_irq_handler() "
				       "> exit on line %d\n",
				       dev->name,__LINE__);
#endif
                                return;
                        }
                        if(!devstat->flag & DEVSTAT_FINAL_STATUS) {
#ifdef FUNCTRACE
                                printk(KERN_INFO "%s: claw_irq_handler() "
				       "> exit on line %d\n",
				       dev->name,__LINE__);
#endif
                                return; 
                        }
                        clear_bit(0, (void *)&p_ch->IO_active);
                        claw_clearbit_busy(TB_RETRY,dev);
                        if (test_and_set_bit(CLAW_BH_ACTIVE,
					     (void *)&p_ch->flag_a) == 0) {
                                queue_task(&p_ch->tq, &tq_immediate);
                                mark_bh(IMMEDIATE_BH);
                        }
#ifdef DEBUGMSG
                        printk(KERN_INFO "%s: process CLAW_STAT_READ exit\n",
			       dev->name);
#endif
#ifdef FUNCTRACE
                        printk(KERN_INFO "%s: claw_irq_handler() "
			       "> exit on line %d\n",dev->name,__LINE__);
#endif
                        return; 
                case CLAW_START_WRITE:
                        if (devstat->dstat & DEV_STAT_UNIT_CHECK) {
                                printk(KERN_INFO "%s: Unit Check Occured "
				       "in Write channel\n",dev->name);
                                clear_bit(0, (void *)&p_ch->IO_active);
                                if (devstat->ii.sense.data[1] & 0x80 ) {
                                        printk(KERN_INFO "%s: Resetting Event "
					       "occurred:\n",dev->name);
                                        init_timer(&p_ch->timer);
                                        p_ch->timer.function =
						(void *)claw_write_retry; 
                                        p_ch->timer.data = (unsigned long)p_ch;
                                        p_ch->timer.expires = jiffies + 10*HZ;
                                        add_timer(&p_ch->timer);
                                        printk(KERN_INFO "%s: write "
					       "connection restarting\n",
					       dev->name);
                                }
#ifdef FUNCTRACE
                                printk(KERN_INFO "%s: claw_irq_handler() "
				       "> exit on line %d\n",
				       dev->name,__LINE__);
#endif
                                return;
                        }
                        if (devstat->dstat & DEV_STAT_UNIT_EXCEP) {
                                        clear_bit(0, (void *)&p_ch->IO_active);
                                        printk(KERN_INFO "%s: Unit Exception "
					       "Occured in Write channel\n",
					       dev->name);
                        }
                        if(!devstat->flag & DEVSTAT_FINAL_STATUS) {
#ifdef FUNCTRACE
                                printk(KERN_INFO "%s: claw_irq_handler() "
				       "> exit on line %d\n",
				       dev->name,__LINE__);
#endif
                                return; 
                        }
                        clear_bit(0, (void *)&p_ch->IO_active);
                        if (claw_test_and_setbit_busy(TB_TX,dev)==0) {
				/* set transmission to busy */
                                claw_write_next(p_ch);
                                claw_clearbit_busy(TB_TX,dev);
                                claw_clear_busy(dev);                   
                        }
                        p_ch_r=(PCHBK)&privptr->channel[READ];
                                 
                        if (test_and_set_bit(CLAW_BH_ACTIVE,
					     (void *)&p_ch_r->flag_a) == 0) {
                                queue_task(&p_ch_r->tq, &tq_immediate);
                                mark_bh(IMMEDIATE_BH);
                        }

#ifdef DEBUGMSG
                        printk(KERN_INFO "%s: process CLAW_STAT_WRITE exit\n",
			       dev->name);
#endif
#ifdef FUNCTRACE
                        printk(KERN_INFO "%s: claw_irq_handler() "
			       "> exit on line %d\n",dev->name,__LINE__);
#endif
                        return;
                default: 
                        printk(KERN_WARNING "%s: wrong selection code - "
			       "irq state=%d\n",dev->name,p_ch->claw_state);
#ifdef FUNCTRACE
                        printk(KERN_INFO "%s: claw_irq_handler() "
			       "> exit on line %d\n",dev->name,__LINE__);
#endif                          
                        return;  
        }

}       /*   end of claw_irq_handler    */


/*-------------------------------------------------------------------*
*       claw_irq_bh                                                  *
*                                                                    *
*--------------------------------------------------------------------*/
static void claw_irq_bh ( PCHBK p_ch)
{

        net_device  *dev;
        PPRIV       privptr;             
        
        dev = (net_device *)p_ch->dev;
#ifdef FUNCTRACE
        printk(KERN_INFO "%s: claw_irq_bh() > enter  \n",dev->name);
        printk(KERN_INFO "%s: variable p_ch =\n",dev->name);
        dumpit((char *) p_ch, sizeof(CHBK));
#endif

        privptr = (PPRIV) dev->priv; 
   
#ifdef DEBUGMSG
        printk(KERN_INFO "%s: bh routine - state-%02x\n" ,dev->name,
	       p_ch->claw_state);
#endif 

        unpack_read(dev);
        clear_bit(CLAW_BH_ACTIVE, (void *)&p_ch->flag_a);
#ifdef FUNCTRACE
        printk(KERN_INFO "%s: claw_irq_bh() > exit on line %d\n",
	       dev->name,__LINE__);
#endif
        return;
}       /*    end of claw_irq_bh    */

/*-------------------------------------------------------------------*
*       claw_release                                                 *
*                                                                    *
*--------------------------------------------------------------------*/
static int claw_release(net_device *dev)
{   
        int                rc;
        int                i;
        __u8               flags = 0x00;
        unsigned long      saveflags;
        unsigned long      parm;
        PPRIV              privptr;
        DECLARE_WAITQUEUE(wait, current);
        PCCWBK             p_this_ccw;
        PCCWBK             p_buf;

	if (!dev) 
                return 0;
        privptr = (PPRIV) dev->priv;     
        if (!privptr)
                return 0;
#ifdef FUNCTRACE
        printk(KERN_INFO "%s: claw_release() > enter  \n",dev->name);
        printk(KERN_INFO "%s: variable dev =\n",dev->name);
        dumpit((char *) dev, sizeof(net_device));
        printk(KERN_INFO "%s: variable privptr=\n",dev->name);
        dumpit((char *) dev->priv, sizeof(CLAW_PRIVBK));
#endif
        privptr->RELEASE_PEND=1;
        claw_setbit_busy(TB_STOP,dev);    
        for ( i = 1; i >=0 ;  i--) {
                s390irq_spin_lock_irqsave(privptr->channel[i].irq, saveflags);
                del_timer(&privptr->channel[READ].timer);  
 		privptr->channel[i].claw_state = CLAW_STOP;
                privptr->channel[i].IO_active = 0;  
                parm = (unsigned long) &privptr->channel[i];
                rc = halt_IO (privptr->channel[i].irq, parm, flags );
	        if (privptr->SYSTEM_VALIDATE_COMP==0x00)  /* never opened? */
                   init_waitqueue_head(&privptr->channel[i].wait);	
                add_wait_queue(&privptr->channel[i].wait, &wait);
                current->state = TASK_INTERRUPTIBLE;
	        s390irq_spin_unlock_irqrestore(privptr->channel[i].irq,
					       saveflags); 
	        schedule();
	        remove_wait_queue(&privptr->channel[i].wait, &wait);
	        if (rc != 0) {
                        ccw_check_return_code(dev, rc);
                }
        }
        free_pages((unsigned long)privptr->p_buffer_ccw,
		   (int)privptr->p_buffer_ccw_num);
        if (privptr->p_env->readSize < PAGE_SIZE) {
                free_pages((unsigned long)privptr->p_buffer_read,
			   (int)privptr->p_buffer_read_num);
        }
        else { 
                p_buf=privptr->p_read_active_first;
                while (p_buf!=NULL) {
                        free_pages((unsigned long)p_buf->p_buffer,
				   (int)privptr->p_buffer_pages_perread );
                        p_buf=p_buf->next;
                }
        }
        if (privptr->p_env->write_size < PAGE_SIZE ) {
                free_pages((unsigned long)privptr->p_buffer_write,
			   (int)privptr->p_buffer_write_num);
        }
        else {
                p_buf=privptr->p_write_active_first;
                while (p_buf!=NULL) {
                        free_pages((unsigned long)p_buf->p_buffer,
				   (int)privptr->p_buffer_pages_perwrite );
                        p_buf=p_buf->next;
                }
        }
        privptr->p_buffer_ccw=NULL;
        privptr->p_buffer_read=NULL;
        privptr->p_buffer_write=NULL;
        privptr->SYSTEM_VALIDATE_COMP=0;
        privptr->RELEASE_PEND=0;
        /*      Remove any writes that were pending and reset all reads   */
        p_this_ccw=privptr->p_read_active_first;
        while (p_this_ccw!=NULL) {
                p_this_ccw->header.length=0xffff;
                p_this_ccw->header.opcode=0xff;
                p_this_ccw->header.flag=0x00;                     
                p_this_ccw=p_this_ccw->next;
        }

        while (privptr->p_write_active_first!=NULL) {
                p_this_ccw=privptr->p_write_active_first;
                p_this_ccw->header.flag=CLAW_PENDING;
                privptr->p_write_active_first=p_this_ccw->next;
                p_this_ccw->next=privptr->p_write_free_chain;
                privptr->p_write_free_chain=p_this_ccw;
                ++privptr->write_free_count;
        }
        privptr->p_write_active_last=NULL;
        privptr->more_to_come_logical_link = -1;
        privptr->MoreToComeSkipping = 1;
        privptr->more_to_come_offset=0;

        if (((privptr->channel[READ].last_dstat |
	      privptr->channel[WRITE].last_dstat) &
             ~(DEV_STAT_CHN_END | DEV_STAT_DEV_END)) != 0x00) {
                printk(KERN_WARNING "%s: channel problems during close - "
		       "read: %02x -  write: %02x\n",
		       dev->name, privptr->channel[READ].last_dstat,
		       privptr->channel[WRITE].last_dstat);
        }
#ifdef  MODULE
        MOD_DEC_USE_COUNT;
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
        dev->start=0;
#endif
#ifdef FUNCTRACE
        printk(KERN_INFO "%s: claw_release() > exit on line %d\n",
	       dev->name,__LINE__);
#endif
        return 0;
}      /* end of claw_release     */

#ifdef CLAW_CHANDEV
/*-------------------------------------------------------------------*
*       claw_chandev_release                                         *
*                                                                    *
*--------------------------------------------------------------------*/
static int claw_chandev_release(net_device *dev)
{   
        int                rc;
        PPRIV              privptr;
        PENV               p_env;
        
#ifdef FUNCTRACE
        printk(KERN_INFO "%s: claw_chandev_release() > enter  \n",dev->name);
        printk(KERN_INFO "%s: variable dev =\n",dev->name);
        dumpit((char *) dev, sizeof(net_device));
#endif
        if (!dev) 
                return 0;
        privptr = (PPRIV) dev->priv;  
        if (!privptr)
                return 0;
        p_env = (PENV) privptr->p_env;
        rc = claw_release(dev); // Call regular release to shutdown the device
        if (privptr->p_more_to_come_envelope!=NULL) {
                kfree(privptr->p_more_to_come_envelope);
                privptr->p_more_to_come_envelope=NULL;
        }
        if (privptr!=NULL) {
                kfree(privptr);
                privptr=NULL;
                dev->priv = NULL;
        }
        if (num_claw_devices > 0) 
            --num_claw_devices;
	if (p_env) {
           memset(p_env, 0, sizeof(CLAW_ENV));
	}
#ifdef FUNCTRACE
        printk(KERN_INFO "%s: claw_chandev_release() "
	       "> exit on line %d\n",dev->name,__LINE__);
#endif
        return 0;
}      /* end of claw_chandev_release     */
#endif 

/*-------------------------------------------------------------------*
*       claw_write_retry                                             *
*                                                                    *
*--------------------------------------------------------------------*/

static void claw_write_retry ( PCHBK p_ch )
{

        net_device  *dev;


#ifdef FUNCTRACE
        printk(KERN_INFO "claw: claw_write_retry() > enter  \n");
        printk(KERN_INFO "claw: variable p_ch =\n");
        dumpit((char *) p_ch, sizeof(CHBK));
#endif

        if (p_ch->claw_state == CLAW_STOP) {
#ifdef FUNCTRACE
                printk(KERN_INFO "claw: claw_write_retry() "
		       "> exit on line %d\n", __LINE__);
#endif

        return;
        }

        dev = (net_device *) p_ch->dev; 
        
#ifdef DEBUGMSG
        printk( KERN_INFO "%s: write retry - state-%02x\n" ,
		dev->name, p_ch->claw_state);
#endif 

        start_output_io( dev );

#ifdef FUNCTRACE
        printk(KERN_INFO "%s: claw_write_retry() "
	       "> exit on line %d\n",dev->name,__LINE__);
#endif
        return;
}      /* end of claw_write_retry      */


/*-------------------------------------------------------------------*
*       claw_write_next                                              *
*                                                                    *
*--------------------------------------------------------------------*/

static void claw_write_next ( PCHBK p_ch )
{

        net_device  *dev;
        PPRIV           privptr=NULL;


#ifdef FUNCTRACE
        printk(KERN_INFO "%s: claw_write_next() > enter  \n",p_ch->dev->name);
        printk(KERN_INFO "%s: variable p_ch =\n",p_ch->dev->name);
        dumpit((char *) p_ch, sizeof(CHBK));
#endif

        if (p_ch->claw_state == CLAW_STOP) {
                return;
        }
         
        dev = (net_device *) p_ch->dev;
	privptr = (PPRIV) dev->priv;
	
        free_write_buffers( dev );

        if (privptr->p_write_active_first!=NULL) {
                start_output_io(dev);
        }


#ifdef FUNCTRACE
        printk(KERN_INFO "%s: claw_write_next() > exit on line %d\n",
	       dev->name,__LINE__);
#endif
        return;
}      /* end of claw_write_next      */

/*-------------------------------------------------------------------*
*                                                                    *
*       claw_timer                                                   *
*--------------------------------------------------------------------*/

static void claw_timer ( PCHBK p_ch )
{
#ifdef FUNCTRACE
        printk(KERN_INFO "%s: claw_timer() > entry\n",p_ch->dev->name);
        printk(KERN_INFO "%s: variable p_ch =\n",p_ch->dev->name);
        dumpit((char *) p_ch, sizeof(CHBK));
#endif

        p_ch->flag |= CLAW_TIMER;
        wake_up(&p_ch->wait);  
#ifdef FUNCTRACE
        printk(KERN_INFO "%s: claw_timer() > exit on line %d\n",
	       p_ch->dev->name, __LINE__);
#endif
        return;
}      /* end of claw_timer  */


/*
*
*       functions
*/

#ifndef CLAW_CHANDEV
/*-------------------------------------------------------------------*
 *  Check CLAW device SID check                                      *
 *-------------------------------------------------------------------*/
static int claw_sid_check(senseid_t *id)
 {
        int rc;
#ifdef FUNCTRACE
        printk(KERN_INFO "claw: claw_sid_check() > enter  \n");
        dumpit((char *)id,sizeof(senseid_t));
#endif

        rc=1;
        if (id->cu_type==0x3088 && id->cu_model==0x61 ) {
                rc=0;
        } 
#ifdef FUNCTRACE
        printk(KERN_INFO "claw: claw_sid_check() "
	       "> exit on line %d, rc = %d\n", __LINE__, rc);
#endif
        return rc;
} /*  end of claw_sid_check  */

/*-------------------------------------------------------------------*
 * check read/write device number and SenseID                        *
 *                                                                   *
 *-------------------------------------------------------------------*/  
static int claw_channel_check(int devno)
{
        int     i;
        int             rc;
        int             temp_devno[2];
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
        dev_info_t temp;
#else   
        s390_dev_info_t temp;
#endif

#ifdef FUNCTRACE
        printk(KERN_INFO "claw: claw_channel_check() > enter  \n");
#endif
        rc=1;

        /* Check read device number is even */
        if ((devno%2)!=0) {
                printk(KERN_INFO "claw: dev %04x is not an even address\n",
		       devno );
#ifdef FUNCTRACE
                printk(KERN_INFO "claw: claw_channel_check() "
		       "> exit on line %d, rc = %d\n", __LINE__, rc);
#endif

                return rc;
        }
        temp_devno[READ]=devno;
        temp_devno[WRITE]=devno+1;

        for (i = 0; i < 2 ; i++) { 
                if (get_irq_by_devno(temp_devno[i]) == -1 ) {
                        printk(KERN_INFO "claw: claw_channel_check(), "
			       "call to get_irq_by_devno() failed, "
			       "devno=%d\n", temp_devno[i]);
#ifdef FUNCTRACE
                        printk(KERN_INFO "claw: claw_channel_check() "
			       "> exit on line %d, rc=ENODEV\n", __LINE__);
#endif                  
                        return -ENODEV;
                }
        }
        /* Check SenseID  for each device */
        for (i = 0; i < 2 ; i++) {
                rc=get_dev_info_by_devno(temp_devno[i], &temp);
                if (rc != 0 ) {
                        printk(KERN_INFO "claw: claw_channel_check(), "
			       "call to get_dev_info() failed, rc=%d, "
			       "devno=%d\n", rc, temp_devno[i]);
#ifdef FUNCTRACE
                        printk(KERN_INFO "claw: claw_channel_check() "
			       "> exit on line %d, rc = %d\n", __LINE__, rc);
#endif                  
                        return -rc;
                }
                rc=claw_sid_check(&temp.sid_data);
                if (rc != 0 ) {
                        printk(KERN_INFO "claw: claw_channel_check(), "
			       "call to claw_sid_check(), failed rc=%d, "
			       "devno=%d\n", rc, temp_devno[i]);
#ifdef FUNCTRACE
                        printk(KERN_INFO "claw: claw_channel_check() "
			       "> exit on line %d, rc=%d\n", __LINE__, rc);
#endif
                        return rc;
                }
        }
        rc=0;
#ifdef FUNCTRACE
        printk(KERN_INFO "claw: claw_channel_check() "
	       "> exit on line %d, rc=0\n", __LINE__);
#endif
        return rc;              
}      /*  end of claw_channel_check   */   
#endif

/*-------------------------------------------------------------------*
*                                                                    *
*     add_claw_reads                                                 *
*                                                                    *
*--------------------------------------------------------------------*/

static int add_claw_reads(net_device *dev, PCCWBK p_first, PCCWBK p_last)
{ 
        PPRIV  privptr;
        ccw1_t  temp_ccw;
        PENDCCW p_end;

#ifdef IOTRACE
        PCCWBK  p_buf;
#endif
#ifdef FUNCTRACE
        printk(KERN_INFO "%s: add_claw_reads() > enter  \n",dev->name);
        printk(KERN_INFO "dev\n");
        dumpit((char *) dev, sizeof(net_device));
        printk(KERN_INFO "p_first\n");
        dumpit((char *) p_first, sizeof(CCWBK));
        printk(KERN_INFO "p_last\n");
        dumpit((char *) p_last, sizeof(CCWBK));
#endif
        privptr = dev->priv;
        p_end = privptr->p_end_ccw;

        /* first CCW and last CCW contains a new set of read channel programs
        *       to apend the running channel programs
        */
        if ( p_first==NULL) {
#ifdef FUNCTRACE
                printk(KERN_INFO "%s: add_claw_reads() "
		       "> exit on line %d\n",dev->name,__LINE__);
#endif
                return 0;
        }
        
        /* set up ending CCW sequence for this segment */
        if (p_end->read1) {
                p_end->read1=0x00;    /*  second ending CCW is now active */
                /*      reset ending CCWs and setup TIC CCWs              */
                p_end->read2_nop2.cmd_code = CCW_CLAW_CMD_READFF;
                p_end->read2_nop2.flags  = CCW_FLAG_SLI | CCW_FLAG_SKIP;
                p_last->RTIC1.cda =(__u32)virt_to_phys(&p_end-> read2_nop1);
                p_last->RTIC2.cda =(__u32)virt_to_phys(&p_end-> read2_nop1);
                p_end->read2_nop2.cda=0;
                p_end->read2_nop2.count=1;
        }
        else {
                p_end->read1=0x01;  /* first ending CCW is now active */
                /*      reset ending CCWs and setup TIC CCWs          */
                p_end->read1_nop2.cmd_code = CCW_CLAW_CMD_READFF;
                p_end->read1_nop2.flags  = CCW_FLAG_SLI | CCW_FLAG_SKIP;
                p_last->RTIC1.cda = (__u32)virt_to_phys(&p_end-> read1_nop1);
                p_last->RTIC2.cda = (__u32)virt_to_phys(&p_end-> read1_nop1);
                p_end->read1_nop2.cda=0;
                p_end->read1_nop2.count=1;
        }

        if ( privptr-> p_read_active_first ==NULL ) {
#ifdef DEBUGMSG
                printk(KERN_INFO "%s: add_claw_reads() "
		       "> p_read_active_frist == NULL \n",dev->name);
                printk(KERN_INFO "%s: add_claw_reads() "
		       "> read active first/last changed \n",dev->name);
#endif
                privptr-> p_read_active_first= p_first;  /* set new first */
                privptr-> p_read_active_last = p_last;   /* set new last  */
        }
        else {

#ifdef DEBUGMSG
                printk(KERN_INFO "%s: add_claw_reads() "
		       "> read in progress \n",dev->name);
#endif
                /* set up TIC ccw  */
                temp_ccw.cda= (__u32)virt_to_phys(&p_first-> read);
                temp_ccw.count=0;
                temp_ccw.flags=0;
                temp_ccw.cmd_code = CCW_CLAW_CMD_TIC;


                if (p_end->read1) {

                        /* first set of CCW's is chained to the new read
                         * chain, so the second set is chained to the
			 * active chain.
                         * Therefore modify the second set to point to the 
			 * new read chain.
                         * set up TIC CCWs
                         * make sure we update the CCW so channel doesn't
			 * fetch it when it's
                         * only half done */
                        memcpy( &p_end->read2_nop2, &temp_ccw ,
				sizeof(ccw1_t));
                        privptr->p_read_active_last->RTIC1.cda=
				(__u32)virt_to_phys(&p_first-> read);
                        privptr->p_read_active_last->RTIC2.cda=
				(__u32)virt_to_phys(&p_first-> read);

                }
                else {
                        /* make sure we update the CCW so channel doesn't
			 * fetch it when it's
                         * only half done */
                        memcpy( &p_end->read1_nop2, &temp_ccw ,sizeof(ccw1_t));
                        privptr->p_read_active_last->RTIC1.cda=
				(__u32)virt_to_phys(&p_first-> read);
                        privptr->p_read_active_last->RTIC2.cda=
				(__u32)virt_to_phys(&p_first-> read);
                }
                /*      chain in new set of blocks */
                privptr->p_read_active_last->next = p_first;
                privptr->p_read_active_last=p_last;
        } /* end of if ( privptr-> p_read_active_first ==NULL)  */
#ifdef IOTRACE
        printk(KERN_INFO "%s: add_claw_reads() >  dump p_last CCW BK \n",
	       dev->name  );
        dumpit((char *)p_last, sizeof(CCWBK));
        printk(KERN_INFO "%s: add_claw_reads() >  dump p_end CCW BK \n",
	       dev->name  );
        dumpit((char *)p_end, sizeof(ENDCCW));

        printk(KERN_INFO "%s: add_claw_reads() >  dump p_first CCW BK \n",
	       dev->name  );
        dumpit((char *)p_first, sizeof(CCWBK));
        printk(KERN_INFO "%s: add_claw_reads() > exit  \n",dev->name);
        printk(KERN_INFO "%s: AddCLawreads()  >  Dump Active CCW chain \n",
	       dev->name  );
        p_buf=privptr->p_read_active_first;
        while (p_buf!=NULL) {
                dumpit((char *)p_buf, sizeof(CCWBK));
                p_buf=p_buf->next;
        }
#endif
#ifdef FUNCTRACE
        printk(KERN_INFO "%s: add_claw_reads > exit on line %d, rc = 0\n",
	       dev->name,__LINE__);
#endif
        return 0;
}    /*     end of add_claw_reads   */

/*-------------------------------------------------------------------*
 *   ccw_check_return_code                                           *
 *                                                                   *
 *-------------------------------------------------------------------*/

static void inline ccw_check_return_code (net_device *dev, int return_code)
{
#ifdef FUNCTRACE
        printk(KERN_INFO "%s: ccw_check_return_code() > enter  \n",dev->name);
        printk(KERN_INFO "variable dev =\n");
        dumpit((char *) dev, sizeof(net_device));
        printk(KERN_INFO "variable return_code = %d\n",return_code);
#endif
        if (return_code != 0) {
                switch (return_code) {
                        case -EBUSY:  
                                printk(KERN_INFO "%s: Busy !\n", dev->name);
                                break;
                        case -ENODEV:
                                printk(KERN_EMERG "%s: Invalid device "
				       "called for IO\n", dev->name);
                                break;
                        case -EIO:
                                printk(KERN_EMERG "%s: Status pending... \n",
				       dev->name);
                                break;
                        default:
                                printk(KERN_EMERG "%s: Unknown error in "
				       "Do_IO %04x\n", 
				       dev->name, return_code);
                }
        }
#ifdef FUNCTRACE
        printk(KERN_INFO "%s: ccw_check_return_code() "
	       "> exit on line %d\n",dev->name,__LINE__);
#endif
}    /*    end of ccw_check_return_code   */

/*-------------------------------------------------------------------*
*       ccw_check_unit_check                                         *
*--------------------------------------------------------------------*/ 

static void inline ccw_check_unit_check (net_device *dev,  devstat_t *devstat )
{

#ifdef FUNCTRACE
        printk(KERN_INFO "%s: ccw_check_unit_check() > enter\n",dev->name);
        printk(KERN_INFO "%s: variable dev =\n",dev->name);
        dumpit((char *) dev, sizeof(net_device));
        printk(KERN_INFO "%s: variable devstat =\n",dev->name);
        dumpit((char *) devstat, sizeof(devstat_t));
#endif

        printk(KERN_INFO "%s: Unit Check with sense byte:0x%02x%02x\n",
                dev->name,devstat->ii.sense.data[0],devstat->ii.sense.data[1]);

        if (devstat->ii.sense.data[0] & 0x40) {
                if (devstat->ii.sense.data[0] & 0x01) { 
                        printk(KERN_WARNING "%s: Interface disconnect or "
			       "Selective reset occurred (remote side)\n",
			       dev->name);
                }
                else { 
                        printk(KERN_WARNING "%s: System reset occured "
			       "(remote side)\n", dev->name);
                }
        }
        else if (devstat->ii.sense.data[0] & 0x20) {
                if (devstat->ii.sense.data[0] & 0x04) {
                        printk(KERN_WARNING "%s: Data-streaming timeout)\n",
			       dev->name);
                }
                else  {
                        printk(KERN_WARNING "%s: Data-transfer parity "
			       "error\n", dev->name);
                }
        }
        else if (devstat->ii.sense.data[0] & 0x10) {
                if (devstat->ii.sense.data[0] & 0x20) {
                        printk(KERN_WARNING "%s: Hardware malfunction "
			       "(remote side)\n", dev->name);
                } 
                else { 
                        printk(KERN_WARNING "%s: read-data parity error "
			       "(remote side)\n", dev->name);
                }
        }

#ifdef FUNCTRACE
        printk(KERN_INFO "%s: ccw_check_unit_check > exit on line %d\n",
	       dev->name,__LINE__);
#endif
}   /*    end of ccw_check_unit_check    */



/*-------------------------------------------------------------------*
* Dump buffer format                                                 *
*                                                                    *
*--------------------------------------------------------------------*/
#ifdef DEBUG
static void dumpit(char* buf, int len)
{

        __u32      ct, sw, rm, dup;
        char       *ptr, *rptr;
        char       tbuf[82], tdup[82];
#if (UTS_MACHINE == s390x)
        char       addr[22];
#else
        char       addr[12];
#endif
        char       boff[12];
        char       bhex[82], duphex[82];
        char       basc[40];
        
        sw  = 0;
        rptr =ptr=buf;
        rm  = 16;
        duphex[0]  = 0x00;
        dup = 0;
        for ( ct=0; ct < len; ct++, ptr++, rptr++ )  {
                if (sw == 0) {
#if (UTS_MACHINE == s390x)
                        sprintf(addr, "%16.16x",(unsigned long)rptr);
#else
                        sprintf(addr, "%8.8X",(__u32)rptr);
#endif
                        sprintf(boff, "%4.4X", (__u32)ct);
                        bhex[0] = '\0';
                        basc[0] = '\0';
                }
                if ((sw == 4) || (sw == 12)) {
                        strcat(bhex, " ");
                }
                if (sw == 8) {
                        strcat(bhex, "  ");
                }
#if (UTS_MACHINE == s390x)
                sprintf(tbuf,"%2.2X", (unsigned long)*ptr);
#else
                sprintf(tbuf,"%2.2X", (__u32)*ptr);
#endif
                tbuf[2] = '\0'; 
                strcat(bhex, tbuf);
                if ((0!=isprint(*ptr)) && (*ptr >= 0x20)) {
                        basc[sw] = *ptr; 
                }
                else {
                        basc[sw] = '.';
                }
                basc[sw+1] = '\0';
                sw++;
                rm--;
                if (sw==16) {
                        if ((strcmp(duphex, bhex)) !=0) {
                                if (dup !=0) {
                                        sprintf(tdup,"Duplicate as above "
						"to %s", addr);
                                        printk( KERN_INFO "               "
						"     --- %s ---\n",tdup);
                                }
                                printk( KERN_INFO "   %s (+%s) : %s  [%s]\n",
					addr, boff, bhex, basc);
                                dup = 0;
                                strcpy(duphex, bhex);
                        }
                        else {
                                dup++;
                        }
                        sw = 0;
                        rm = 16;
                }
        }  /* endfor */
        
        if (sw != 0) {
                for ( ; rm > 0; rm--, sw++ ) {
                        if ((sw==4) || (sw==12)) strcat(bhex, " ");
                        if (sw==8)               strcat(bhex, "  ");
                        strcat(bhex, "  ");
                        strcat(basc, " ");
                }
                if (dup !=0) {
                        sprintf(tdup,"Duplicate as above to %s", addr);
                        printk( KERN_INFO "               "
				"     --- %s ---\n",tdup);
                }
                printk( KERN_INFO "   %s (+%s) : %s  [%s]\n",
			addr, boff, bhex, basc);
        }
        else {
                if (dup >=1) {
                        sprintf(tdup,"Duplicate as above to %s", addr);
                        printk( KERN_INFO "               "
				"     --- %s ---\n",tdup);
                }
                if (dup !=0) {
                        printk( KERN_INFO "   %s (+%s) : %s  [%s]\n",
				addr, boff, bhex, basc);
                }
        }
        return;

}   /*   end of dumpit  */
#endif

/*-------------------------------------------------------------------*
*       Extract Device id                                            *
*                                                                    *
*--------------------------------------------------------------------*/
static int inline extract_dev_id(char *name)
{
        int     idnum,i;
#ifdef FUNCTRACE
        printk(KERN_INFO "%s: extract_dev_id() > enter\n",name);
        printk(KERN_INFO "%s: name = %s\n",name,name);
#endif
        idnum=0;
        for (i=4;i< strlen(name); i++)  // get the clawxxx to int 
        {
                switch (name[i]) {
                        case '0'...'9':
                                idnum = idnum*10+(name[i]-'0');
                                break;
                        default:
                                idnum = -1;  // bad devid 
                }
        }
        
#ifdef FUNCTRACE
        printk(KERN_INFO "%s: extract_dev_id() > exit on line %d, rc= %d\n",
	       name, __LINE__,idnum);
#endif

        return (idnum);
                
} /*  end of extract_dev_id  */

/*-------------------------------------------------------------------*
*               find_link                                             *
*--------------------------------------------------------------------*/
static int find_link(net_device *dev, char *host_name, char *ws_name )
{
        PPRIV           privptr;
        PENV            p_env;

#ifdef FUNCTRACE
        printk(KERN_INFO "%s: find_link() > enter  \n",dev->name);
        printk(KERN_INFO "%s: variable dev = \n",dev->name);
        dumpit((char *) dev, sizeof(net_device));
        printk(KERN_INFO "%s: variable host_name = %s\n",dev->name, host_name);
        printk(KERN_INFO "%s: variable ws_name = %s\n",dev->name, ws_name);   
#endif
        privptr=dev->priv;
        p_env=privptr->p_env;


        if ( memcmp(HOST_APPL_NAME, host_name, 8)!=0 ||
	     memcmp(p_env->ApiType , ws_name, 8)!=0 ) {
                return EINVAL;
        }

#ifdef FUNCTRACE
        printk(KERN_INFO "%s: find_link() > exit on line %d\n",
	       dev->name,__LINE__);
#endif   
        return 0;
}    /*    end of find_link    */

/*-------------------------------------------------------------------*
 *   claw_hw_tx                                                      *
 *                                                                   *
 *                                                                   *
 *-------------------------------------------------------------------*/

static int claw_hw_tx(struct sk_buff *skb, net_device *dev, long linkid)
{
        int                             rc=0;
        PPRIV                           privptr;
        PCCWBK                          p_this_ccw;
        PCCWBK                          p_first_CCW;
        PCCWBK                          p_last_CCW;
        __u32                           numBuffers;
        signed long                     length_of_data;
        unsigned long                   bytes_in_this_buffer;
        unsigned char                   *p_data_address;
        PENDCCW                         p_end;
        ccw1_t                          temp_CCW;
        PCHBK                           p_ch;
        int                             lock;
#ifdef IOTRACE
        PCCWBK                          p_buf;
#endif
#ifdef FUNCTRACE
        printk(KERN_INFO "%s: claw_hw_tx() > enter\n",dev->name);
        printk(KERN_INFO "%s: variable dev skb =\n",dev->name);
        dumpit((char *) skb, sizeof(struct sk_buff));
        printk(KERN_INFO "%s: variable dev =\n",dev->name);
        dumpit((char *) dev, sizeof(net_device));
        printk(KERN_INFO "%s: variable linkid = %ld\n",dev->name,linkid);
#endif

   
        privptr = (PPRIV) (dev->priv);
        p_ch=(PCHBK)&privptr->channel[WRITE];

#ifdef IOTRACE
        printk(KERN_INFO "%s: claw_hw_tx() > dump sk_buff  \n",dev->name);
        dumpit((char *)skb ,sizeof(struct sk_buff));
#endif
	free_write_buffers(dev);   // Clean up free chain if posible

        /*  scan the write queue to free any completed write packets   */
        p_first_CCW=NULL;
        p_last_CCW=NULL;

        if (linkid!=0 ) {
                if (claw_check_busy(dev)) {
                        if (privptr->write_free_count!=0) {
                                claw_clear_busy(dev);
                        }
                        else { 
                                start_output_io(dev );
                                free_write_buffers( dev );
				// OK try to free again
                                if (privptr->write_free_count!=0) {
#ifdef IOTRACE
					printk(KERN_INFO "%s:  busy  "
					       "(claw_check_busy) no free "
					       "write buffers\n", dev->name);
#endif
	   				rc=-EBUSY;
					goto Done;
                                }
                                else {
					claw_clear_busy(dev);        
                                }
                        }
                } 
                /*  tx lock  */
                if (claw_test_and_setbit_busy(TB_TX,dev)) {
			/* set transmission to busy */
#ifdef DEBUGMSG
                        printk(KERN_INFO "%s:  busy  "
			       "(claw_test_and_setbit_busy)\n", dev->name);
                        start_output_io(dev );
#endif
                        rc=-EBUSY;
                        goto Done;
                }
        }


        /* See how many write buffers are required to hold this data */
        numBuffers= ( skb->len + privptr->p_env->write_size - 1) /
		( privptr->p_env->write_size);
                
        /* If that number of buffers isn't available, give up for now */
        if (privptr->write_free_count < numBuffers || 
            privptr->p_write_free_chain == NULL ) {
                
                claw_setbit_busy(TB_NOBUFFER,dev);

#ifdef DEBUGMSG
                printk(KERN_INFO "%s:  busy  "
		       "(claw_setbit_busy(TB_NOBUFFER))\n", dev->name);
                printk(KERN_INFO "       free_count: %d, numBuffers : %d\n",
		       (int)privptr->write_free_count,(int) numBuffers );
#endif            
                rc=-EBUSY;
                goto Done2;
        }
                
        p_data_address=skb->data;
        length_of_data=skb->len;

        while (length_of_data > 0) {

#ifdef DEBUGMSG
                printk(KERN_INFO "%s: claw_hw_tx() "
		       "> length-of-data is %ld \n",dev->name ,length_of_data);
                dumpit((char *)p_data_address ,64);
#endif
                p_this_ccw=privptr->p_write_free_chain;       // get a block
                privptr->p_write_free_chain=p_this_ccw->next; // update chain
		                                            // pointer
                p_this_ccw->next=NULL;
                --privptr->write_free_count;                // -1
                if ((length_of_data) <= (privptr->p_env->write_size)) {
                        bytes_in_this_buffer=length_of_data;
                } 
                else {
                        bytes_in_this_buffer=privptr->p_env->write_size;
                }
                memcpy( p_this_ccw->p_buffer,p_data_address,
			bytes_in_this_buffer);
                length_of_data-=bytes_in_this_buffer;
                p_data_address+=(unsigned long)bytes_in_this_buffer;
                /*      setup Write CCW         */
                p_this_ccw->Write.cmd_code = (linkid * 8) +1;
                if (length_of_data>0) {
                        p_this_ccw->Write.cmd_code+=MORE_to_COME_FLAG;
                }
                p_this_ccw->Write.count=bytes_in_this_buffer;
                /*      now add to end of this chain    */
                if (p_first_CCW==NULL)    {
                        p_first_CCW=p_this_ccw;
                }
                if (p_last_CCW!=NULL) {
                        p_last_CCW->next=p_this_ccw;
                        /*      set up TIC ccws         */
                        p_last_CCW->WTIC1.cda=
				(__u32)virt_to_phys(&p_this_ccw->Write);
                }
                p_last_CCW=p_this_ccw;      // save new last block
#ifdef IOTRACE
		printk(KERN_INFO "%s: claw_hw_tx() > CCW and Buffer "
		       "%ld bytes long \n",dev->name,bytes_in_this_buffer);
                dumpit((char *)p_this_ccw, sizeof(CCWBK));
                dumpit((char *)p_this_ccw->p_buffer, 64);
#endif          
        }

        /*      FirstCCW and LastCCW now contain a new set of write channel
        *       programs to append to the running channel program
        */

        if (p_first_CCW!=NULL) {
                /*      setup ending ccw sequence for this segment */
                p_end=privptr->p_end_ccw;
                if (p_end->write1) {
                        p_end->write1=0x00;  // second end ccw is now active
                        /*      set up Tic CCWs         */
                        p_last_CCW->WTIC1.cda=
				(__u32)virt_to_phys(&p_end->write2_nop1);
                        p_end->write2_nop2.cmd_code = CCW_CLAW_CMD_READFF;
                        p_end->write2_nop2.flags    = CCW_FLAG_SLI |
				CCW_FLAG_SKIP;
                        p_end->write2_nop2.cda=0;
                        p_end->write2_nop2.count=1;
                }
                else {  /*  end of if (p_end->write1)*/ 
                        p_end->write1=0x01;      // first end ccw is now active
                        /*      set up Tic CCWs         */
                        p_last_CCW->WTIC1.cda=
				(__u32)virt_to_phys(&p_end->write1_nop1);
                        p_end->write1_nop2.cmd_code = CCW_CLAW_CMD_READFF;
                        p_end->write1_nop2.flags    = CCW_FLAG_SLI |
				CCW_FLAG_SKIP;
                        p_end->write1_nop2.cda=0;
                        p_end->write1_nop2.count=1;                       
                }  /* end if if (p_end->write1) */
        
                   
                if (privptr->p_write_active_first==NULL ) {
			privptr->p_write_active_first=p_first_CCW; 
                        privptr->p_write_active_last=p_last_CCW;
                }
                else {

                        /*      set up Tic CCWs         */
                        temp_CCW.cda=(__u32)virt_to_phys(&p_first_CCW->Write);
                        temp_CCW.count=0;
                        temp_CCW.flags=0;
                        temp_CCW.cmd_code=CCW_CLAW_CMD_TIC;

                        if (p_end->write1) {

                                /*
                                * first set of ending CCW's is chained to
				* the new write
                                * chain, so the second set is chained to
				* the active chain
                                * Therefore modify the second set to point
				* the new write chain.
                                */
                                /* make sure we update the CCW atomically
                                * so channel does not fetch it when it's
				* only half done
                                */

                                memcpy( &p_end->write2_nop2, &temp_CCW ,
					sizeof(ccw1_t));         
                                privptr->p_write_active_last->WTIC1.cda=
					(__u32)virt_to_phys(&p_first_CCW->
							    Write);
                        } 
                        else {
                                 
                                /*      make sure we update the CCW atomically
                                *       so channel does not fetch it when
				*       it's only half done
                                */
                                memcpy(&p_end->write1_nop2, &temp_CCW ,
				       sizeof(ccw1_t));
                                privptr->p_write_active_last->WTIC1.cda=
					(__u32)virt_to_phys(&p_first_CCW->
							    Write);

                        } /* end if if (p_end->write1) */ 

                        privptr->p_write_active_last->next=p_first_CCW;
                        privptr->p_write_active_last=p_last_CCW;  
                } 

        } /* endif (p_first_CCW!=NULL)  */


#ifdef IOTRACE
        printk(KERN_INFO "%s: hw_tx() >  Dump Active CCW chain \n",dev->name);
        p_buf=privptr->p_write_active_first;
        while (p_buf!=NULL) {
                dumpit((char *)p_buf, sizeof(CCWBK));
                p_buf=p_buf->next;
        }

        p_buf=(PCCWBK)privptr->p_end_ccw;
        dumpit((char *)p_buf, sizeof(ENDCCW));
#endif


        dev_kfree_skb(skb);

        if (linkid==0) {
                lock=LOCK_NO;
        }
        else  {
                lock=LOCK_YES;
        }
        start_output_io(dev );


        /*      if write free count is zero , set NOBUFFER  */
#ifdef DEBUGMSG
        printk(KERN_INFO "%s: claw_hw_tx() > free_count is %d\n",
	       dev->name, (int) privptr->write_free_count );
#endif

        if (privptr->write_free_count==0) {
                claw_setbit_busy(TB_NOBUFFER,dev);
        }
Done2:
        claw_clearbit_busy(TB_TX,dev);
Done:


#ifdef FUNCTRACE
        printk(KERN_INFO "%s: claw_hw_tx() > exit on line %d, rc = %d \n",
	       dev->name, __LINE__, rc);
#endif
        return(rc);
}    /*    end of claw_hw_tx    */

/*-------------------------------------------------------------------*
*                                                                    *
*     init_ccw_bk                                                    *
*                                                                    *
*--------------------------------------------------------------------*/

static int init_ccw_bk(net_device *dev)
{

        __u32   ccw_blocks_required;
        __u32   ccw_blocks_perpage;
        __u32   ccw_pages_required;
        __u32   claw_reads_perpage=1;
        __u32   claw_read_pages;
        __u32   claw_writes_perpage=1;
        __u32   claw_write_pages;
        void    *p_buffer=NULL;
        PCCWBK          p_free_chain,p_buf,p_last_CCWB,p_first_CCWB;
        PENDCCW         p_endccw=NULL;
        addr_t          real_address;
        PPRIV           privptr=dev->priv;
        PCLAWH          pClawH=NULL;
        addr_t          real_TIC_address;

        int i,j;
#ifdef FUNCTRACE
        printk(KERN_INFO "%s: init_ccw_bk() > enter  \n",dev->name);
        printk(KERN_INFO "%s: variable dev =\n",dev->name);
        dumpit((char *) dev, sizeof(net_device));
#endif

        /*  initialize  statistics field */
        privptr->ActiveLinkID=0;
        /*  initialize  ccwbk pointers  */
        privptr->p_write_free_chain=NULL;     // pointer to free ccw chain
        privptr->p_write_active_first=NULL;   // pointer to the first write ccw
        privptr->p_write_active_last=NULL;    // pointer to the last write ccw
        privptr->p_read_active_first=NULL;    // pointer to the first read ccw
        privptr->p_read_active_last=NULL;     // pointer to the last read ccw
        privptr->p_end_ccw=NULL;              // pointer to ending ccw
        privptr->pClawSignalBlock=NULL;       // pointer to signal block
        memset(&privptr->EndCCW, 0x00, sizeof(ENDCCW));
        memset(&privptr->ctl_bk, 0x00, sizeof(CLAWCTL));

        /*  initialize  free write ccwbk counter  */    
        privptr->write_free_count=0;          // number of free bufs on
	                                      // write free chain
        p_last_CCWB = NULL;
        p_first_CCWB= NULL;

        /* 
        *  we need 1 CCW block for each read buffer, 1 for each    
        *  write buffer, plus 1 for ClawSignalBlock   
        */

        ccw_blocks_required=privptr->p_env->readBuffers+
		privptr->p_env->WriteBuffers+1;

#ifdef DEBUGMSG
        printk(KERN_INFO "%s: init_ccw_bk() > ccw_blocks_required=%d\n",
	       dev->name, ccw_blocks_required);
        printk(KERN_INFO "%s: init_ccw_bk() > PAGE_SIZE=0x%x\n",
	       dev->name, (unsigned int)PAGE_SIZE);
        printk(KERN_INFO "%s: init_ccw_bk() > PAGE_MASK=0x%x\n",
	       dev->name, (unsigned int)PAGE_MASK);
#endif
        /*
        * compute number of CCW blocks that will fit in a page    
        */
        ccw_blocks_perpage= PAGE_SIZE / CCWBK_SIZE;
        ccw_pages_required= (ccw_blocks_required+ccw_blocks_perpage -1) /
		ccw_blocks_perpage;

#ifdef DEBUGMSG
        printk(KERN_INFO "%s: init_ccw_bk() > ccw_blocks_perpage=%d\n",
	       dev->name, ccw_blocks_perpage);
        printk(KERN_INFO "%s: init_ccw_bk() > ccw_pages_required=%d\n",
	       dev->name, ccw_pages_required);
#endif  
        /*
         *  read and write sizes are set by a constant at the begining of the
	 *  module.  Values other than 4k are 
         *  not going to provide good performance.  Until packing buffers
	 *  are supported, values greater
         *  than 4k are a pointless waste of storage.  Less than 4k builds
	 *  channel busy.     
         */ 
        if (privptr->p_env->readSize < PAGE_SIZE) {
	       	claw_reads_perpage= PAGE_SIZE / privptr->p_env->readSize;
		claw_read_pages= (privptr->p_env->readBuffers +
		    		  claw_reads_perpage -1) / claw_reads_perpage;
		if (claw_read_pages > 9) { claw_read_pages = 9; }
         }
         else {       /* > or equal  */
		 privptr->p_buffer_pages_perread=
			 (privptr->p_env->readSize + PAGE_SIZE - 1) / PAGE_SIZE;
		 if (privptr->p_buffer_pages_perread > 9) {
			 privptr->p_buffer_pages_perread = 9;
		 }
		 claw_read_pages= privptr->p_env->readBuffers *
			 privptr->p_buffer_pages_perread;
         }   
        if (privptr->p_env->write_size < PAGE_SIZE) {
		claw_writes_perpage= PAGE_SIZE / privptr->p_env->write_size;
		claw_write_pages= (privptr->p_env->WriteBuffers +
				   claw_writes_perpage -1) /
			claw_writes_perpage;
		if (claw_write_pages > 9 ) {
			claw_write_pages = 9;
		}
        }
        else {      /* >  or equal  */
		privptr->p_buffer_pages_perwrite= (privptr->p_env->readSize +
						  PAGE_SIZE - 1) / PAGE_SIZE;
		if (privptr->p_buffer_pages_perwrite > 9) {
			privptr->p_buffer_pages_perwrite = 9;
		}
		claw_write_pages= privptr->p_env->WriteBuffers *
			privptr->p_buffer_pages_perwrite;
        } 
        
#ifdef DEBUGMSG
        if (privptr->p_env->readSize < PAGE_SIZE) {
		printk(KERN_INFO "%s: init_ccw_bk() > reads_perpage=%d\n",
		       dev->name, claw_reads_perpage);
        } 
        else {
            printk(KERN_INFO "%s: init_ccw_blk() > pages_perread=%d\n",
		   dev->name, privptr->p_buffer_pages_perread);
        }
        printk(KERN_INFO "%s: init_ccw_bk() > read_pages=%d\n",
	       dev->name, claw_read_pages);
        if (privptr->p_env->write_size < PAGE_SIZE) {
		printk(KERN_INFO "%s: init_ccw_bk() > writes_perpage=%d\n",
		       dev->name, claw_writes_perpage);
        }
        else {
		printk(KERN_INFO "%s: init_ccw_bk() > pages_perwrite=%d\n",
		       dev->name, privptr->p_buffer_pages_perwrite);
        }
        printk(KERN_INFO "%s: init_ccw_bk() > write_pages=%d\n",
	       dev->name, claw_write_pages);
#endif


        /*
        *               allocate ccw_pages_required
        */

        if (privptr->p_buffer_ccw==NULL) {
                privptr->p_buffer_ccw=
			(void *)__get_free_pages(GFP_KERNEL|GFP_DMA,
						 ccw_pages_required );
                if (privptr->p_buffer_ccw==NULL) {
                        printk(KERN_INFO "%s: init_ccw_bk() "
			       "> __get_free_pages for CCWs failed : pages "
			       "is %d\n",dev->name,ccw_pages_required );
#ifdef FUNCTRACE
                        printk(KERN_INFO "%s: init_ccw_bk() "
			       "> exit on line %d, rc = ENOMEM\n",
			       dev->name, __LINE__);
#endif
                        return -ENOMEM;
                }
                privptr->p_buffer_ccw_num=ccw_pages_required;
        } 
        memset(privptr->p_buffer_ccw, 0x00, ccw_pages_required * PAGE_SIZE);

        /*
        *               obtain ending ccw block address
        *
        */
         
        privptr->p_end_ccw = (PENDCCW)&privptr->EndCCW; 
        real_address  = (__u32)virt_to_phys(privptr->p_end_ccw);

        /*          Initialize ending CCW block       */
#ifdef DEBUGMSG
        printk(KERN_INFO "%s: init_ccw_bk() "
	       ">  begin initialize ending CCW blocks\n",dev->name);
#endif

        p_endccw=privptr->p_end_ccw;
        p_endccw->real=real_address;
        p_endccw->write1=0x00;
        p_endccw->read1=0x00;

        /*      write1_nop1                                     */
        p_endccw->write1_nop1.cmd_code = CCW_CLAW_CMD_NOP;
        p_endccw->write1_nop1.flags       = CCW_FLAG_SLI | CCW_FLAG_CC;
        p_endccw->write1_nop1.count       = 1;
        p_endccw->write1_nop1.cda         = 0;

        /*      write1_nop2                                     */
        p_endccw->write1_nop2.cmd_code = CCW_CLAW_CMD_READFF;
        p_endccw->write1_nop2.flags        = CCW_FLAG_SLI | CCW_FLAG_SKIP;
        p_endccw->write1_nop2.count      = 1;
        p_endccw->write1_nop2.cda        = 0;

        /*      write2_nop1                                     */
        p_endccw->write2_nop1.cmd_code = CCW_CLAW_CMD_NOP;
        p_endccw->write2_nop1.flags        = CCW_FLAG_SLI | CCW_FLAG_CC;
        p_endccw->write2_nop1.count        = 1;
        p_endccw->write2_nop1.cda          = 0;

        /*      write2_nop2                                     */
        p_endccw->write2_nop2.cmd_code = CCW_CLAW_CMD_READFF;
        p_endccw->write2_nop2.flags        = CCW_FLAG_SLI | CCW_FLAG_SKIP;
        p_endccw->write2_nop2.count       = 1;
        p_endccw->write2_nop2.cda          = 0;

        /*      read1_nop1                                      */
        p_endccw->read1_nop1.cmd_code = CCW_CLAW_CMD_NOP;
        p_endccw->read1_nop1.flags        = CCW_FLAG_SLI | CCW_FLAG_CC;
        p_endccw->read1_nop1.count        = 1;
        p_endccw->read1_nop1.cda          = 0;

        /*      read1_nop2                                      */
        p_endccw->read1_nop2.cmd_code = CCW_CLAW_CMD_READFF;
        p_endccw->read1_nop2.flags        = CCW_FLAG_SLI | CCW_FLAG_SKIP;
        p_endccw->read1_nop2.count        = 1;
        p_endccw->read1_nop2.cda          = 0;

        /*      read2_nop1                                      */
        p_endccw->read2_nop1.cmd_code = CCW_CLAW_CMD_NOP;
        p_endccw->read2_nop1.flags        = CCW_FLAG_SLI | CCW_FLAG_CC;
        p_endccw->read2_nop1.count        = 1;
        p_endccw->read2_nop1.cda          = 0;

        /*      read2_nop2                                      */
        p_endccw->read2_nop2.cmd_code = CCW_CLAW_CMD_READFF;
        p_endccw->read2_nop2.flags        = CCW_FLAG_SLI | CCW_FLAG_SKIP;
        p_endccw->read2_nop2.count        = 1;
        p_endccw->read2_nop2.cda          = 0;

        
#ifdef IOTRACE
        printk(KERN_INFO "%s: init_ccw_bk() >  dump claw ending CCW BK \n",
	       dev->name  );
        dumpit((char *)p_endccw, sizeof(ENDCCW));
#endif

        /*
        *                               Build a chain of CCWs
        *
        */

#ifdef DEBUGMSG
        printk(KERN_INFO "%s: init_ccw_bk() "
	       ">  Begin build a chain of CCW buffer \n",dev->name  );
#endif
        p_buffer=privptr->p_buffer_ccw;

        p_free_chain=NULL;
        for (i=0 ; i < ccw_pages_required; i++ ) {
                real_address  = (__u32)virt_to_phys(p_buffer);
                p_buf=p_buffer;
                for (j=0 ; j < ccw_blocks_perpage ; j++) {
                        p_buf->next  = p_free_chain;
                        p_free_chain = p_buf;
                        p_buf->real=(__u32)virt_to_phys(p_buf);
                        ++p_buf;
                }
                p_buffer+=PAGE_SIZE;
        }

#ifdef DEBUGMSG
        printk(KERN_INFO "%s: init_ccw_bk() "
	       ">  End build a chain of CCW buffer \n",dev->name  );
        p_buf=p_free_chain;
        while (p_buf!=NULL) {
                dumpit((char *)p_buf, sizeof(CCWBK));
                p_buf=p_buf->next;
        }
#endif

        /*
        *                               Initialize ClawSignalBlock
        *
        */
#ifdef DEBUGMSG
        printk(KERN_INFO "%s: init_ccw_bk() "
	       ">  Begin initialize ClawSignalBlock \n",dev->name );
#endif
        if (privptr->pClawSignalBlock==NULL) {
                privptr->pClawSignalBlock=p_free_chain;
                p_free_chain=p_free_chain->next;
                pClawH=(PCLAWH)privptr->pClawSignalBlock;
                pClawH->length=0xffff;
                pClawH->opcode=0xff;
                pClawH->flag=CLAW_BUSY;
        }
#ifdef DEBUGMSG
        printk(KERN_INFO "%s: init_ccw_bk() "
	       ">  End initialize ClawSignalBlock\n",dev->name  );
        dumpit((char *)privptr->pClawSignalBlock, sizeof(CCWBK));
#endif

        /*
        *               allocate write_pages_required and add to free chain
        */
        if (privptr->p_buffer_write==NULL) {
		if (privptr->p_env->write_size < PAGE_SIZE) {
			privptr->p_buffer_write=
				(void *)__get_free_pages(GFP_KERNEL|GFP_DMA,
							 claw_write_pages );
			if (privptr->p_buffer_write==NULL) {
				printk(KERN_INFO "claw: init_ccw_bk() "
				       "> __get_free_pages for writes buf"
				       "failed : get is for %d pages\n",
				       claw_write_pages );
				free_pages((unsigned long)privptr->p_buffer_ccw,
					   (int)privptr->p_buffer_ccw_num);
				privptr->p_buffer_ccw=NULL;
#ifdef FUNCTRACE
				printk(KERN_INFO "%s: init_ccw_bk() "
				       "> exit on line %d, rc = ENOMEM\n",
				       dev->name, __LINE__);
#endif
				return -ENOMEM;
			}
			/*
			 *           Build CLAW write free chain
			 *
			 */

			memset(privptr->p_buffer_write, 0x00,
			       ccw_pages_required * PAGE_SIZE);
#ifdef DEBUGMSG
			printk(KERN_INFO "%s: init_ccw_bk() "
			       ">  Begin build claw write free chain \n",
			       dev->name  );
#endif
			privptr->p_write_free_chain=NULL;

			p_buffer=privptr->p_buffer_write;

			for (i=0 ; i< privptr->p_env->WriteBuffers ; i++) {
				//  get a CCW
				p_buf        = p_free_chain;
				p_free_chain = p_buf->next;
				p_buf->next  =privptr->p_write_free_chain;
				privptr->p_write_free_chain = p_buf;
				// save virtual
				p_buf->p_buffer=(PCLAWBUF)p_buffer;
				// buffer address
				p_buf-> Write.cda        =
					(__u32)virt_to_phys(p_buffer);
				p_buf-> Write.flags      =
					CCW_FLAG_SLI | CCW_FLAG_CC;
				p_buf-> WreadFF.cmd_code =
					CCW_CLAW_CMD_READFF;
				p_buf-> WreadFF.flags    =
					CCW_FLAG_SLI | CCW_FLAG_CC;
				p_buf-> WreadFF.count    = 1;
				p_buf-> WreadFF.cda      =
					(__u32)virt_to_phys(&p_buf-> header.flag);
				p_buf-> WTIC1.cmd_code = CCW_CLAW_CMD_TIC;
				p_buf-> WTIC1.flags      = 0;
				p_buf-> WTIC1.count      = 0;
				
				if (((unsigned long)p_buffer+privptr->
				     p_env->write_size)
				    >= ((unsigned long)(p_buffer+2*
							(privptr->p_env->
							 write_size) -1) &
					PAGE_MASK  ) ) {
					p_buffer= p_buffer+privptr->p_env->
						write_size;
				}               
			}
		}
		else      /*  Buffers are PAGE_SIZE or bigger.
		     	      Allocate one buffer per get_free_pages     */
		{
			privptr->p_write_free_chain=NULL;
			
			for (i = 0; i< privptr->p_env->WriteBuffers ; i++) {
				p_buffer=(void *)__get_free_pages
					(GFP_KERNEL|GFP_DMA,
					 privptr->p_buffer_pages_perwrite );
#ifdef IOTRACE   
		     		printk(KERN_INFO "%s: init_ccw_bk() "
				       "> __get_free_pages for writes buf: "
				       "get for %ld pages at %x\n",
				       dev->name,
				       privptr->p_buffer_pages_perwrite,
				       p_buffer);
#endif
     				if (p_buffer==NULL) {
					printk(KERN_INFO "%s: init_ccw_bk() "
					       "> __get_free_pages for "
					       "writes buf failed : get "
					       "is for %d pages\n",
					       dev->name,privptr->
					       p_buffer_pages_perwrite);
     					free_pages((unsigned long)privptr->
						   p_buffer_ccw,
						   (int)privptr->
						   p_buffer_ccw_num);
     					privptr->p_buffer_ccw=NULL;
#ifdef FUNCTRACE
     					printk(KERN_INFO "%s: init_ccw_bk() "
					       "> exit on line %d, "
					       "rc = ENOMEM\n",
					       dev->name,__LINE__);
#endif
     					return -ENOMEM;
     				}  /* Error on get_pages   */
     				memset(p_buffer, 0x00,
				       privptr->p_env->write_size );
     				p_buf         = p_free_chain;
     				p_free_chain  = p_buf->next;
     				p_buf->next   = privptr->p_write_free_chain;
     				privptr->p_write_free_chain = p_buf;
     				privptr->p_buffer_write = p_buf;
				// save virtual buffer address
     				p_buf->p_buffer=(PCLAWBUF)p_buffer;
     				p_buf-> Write.cda     =
					(__u32)virt_to_phys(p_buffer);
     				p_buf-> Write.flags   =
					CCW_FLAG_SLI | CCW_FLAG_CC;
     				p_buf-> WreadFF.cmd_code = CCW_CLAW_CMD_READFF;
     				p_buf-> WreadFF.flags = 
					CCW_FLAG_SLI | CCW_FLAG_CC;
     				p_buf-> WreadFF.count = 1;
     				p_buf-> WreadFF.cda = 
					(__u32)virt_to_phys(&p_buf->
							    header.flag);
     				p_buf-> WTIC1.cmd_code = CCW_CLAW_CMD_TIC;
     				p_buf-> WTIC1.flags   = 0;
     				p_buf-> WTIC1.count   = 0;
     			}  /* for all WriteBuffers   */
     			
     		}    /* else buffers are PAGE_SIZE or bigger */
	   
	}
        privptr->p_buffer_write_num=claw_write_pages;
        privptr->write_free_count=privptr->p_env->WriteBuffers;
        

#ifdef DEBUGMSG
        printk(KERN_INFO "%s: init_ccw_bk() "
	       ">  End build claw write free chain \n",dev->name );
        p_buf=privptr->p_write_free_chain;
        while (p_buf!=NULL) {
                dumpit((char *)p_buf, sizeof(CCWBK));
                p_buf=p_buf->next;
        }
#endif
        /*
        *               allocate read_pages_required and chain to free chain
        */
        if (privptr->p_buffer_read==NULL) {
            if (privptr->p_env->readSize < PAGE_SIZE)  {
                privptr->p_buffer_read=
			(void *)__get_free_pages(GFP_KERNEL|GFP_DMA,
						 claw_read_pages );
                if (privptr->p_buffer_read==NULL) {
                        printk(KERN_INFO "%s: init_ccw_bk() "
			       "> __get_free_pages for read buf failed : "
			       "get is for %d pages\n", 
                                dev->name,claw_read_pages );
                        free_pages((unsigned long)privptr->p_buffer_ccw,
				   (int)privptr->p_buffer_ccw_num);
                        if (privptr->p_env->write_size < PAGE_SIZE) {
                                free_pages((unsigned long)privptr->
					   p_buffer_write,
					   (int)privptr->p_buffer_write_num);
                        }
                        else {
                                p_buf=privptr->p_buffer_write;
                                while (p_buf!=NULL) {
                                        free_pages((unsigned long)p_buf->
						   p_buffer,
						   (int)privptr->
						   p_buffer_pages_perwrite );
                                        p_buf=p_buf->next;
                                }
                        }
                        privptr->p_buffer_ccw=NULL;
                        privptr->p_buffer_write=NULL;
#ifdef FUNCTRACE
                        printk(KERN_INFO "%s: init_ccw_bk() "
			       "> exit on line %d, rc = ENOMEM\n",
			       dev->name, __LINE__);
#endif                  
                        return -ENOMEM;
                }
                memset(privptr->p_buffer_read,0x00,claw_read_pages*PAGE_SIZE);
                privptr->p_buffer_read_num=claw_read_pages;
                /*
                *                               Build CLAW read free chain
                *
                */
#ifdef DEBUGMSG
                printk(KERN_INFO "%s: init_ccw_bk() "
		       ">  Begin build claw read free chain \n",dev->name  );
#endif
                p_buffer=privptr->p_buffer_read;

                for (i=0 ; i< privptr->p_env->readBuffers ; i++) {
                        p_buf        = p_free_chain;
                        p_free_chain = p_buf->next;

                        if (p_last_CCWB==NULL) {
                                p_buf->next=NULL;
                                real_TIC_address=0;
                                p_last_CCWB=p_buf;
                        }
                        else {
                                p_buf->next=p_first_CCWB;
                                real_TIC_address=
					(__u32)virt_to_phys(&p_first_CCWB->
							    read );
                        }

                        p_first_CCWB=p_buf;

			// save virtual buffer address
                        p_buf->p_buffer=(PCLAWBUF)p_buffer;      
                        /*  initialize read command */
                        p_buf-> read.cmd_code = CCW_CLAW_CMD_READ;
                        p_buf-> read.cda = (__u32)virt_to_phys(p_buffer);
                        p_buf-> read.flags = CCW_FLAG_SLI | CCW_FLAG_CC;
                        p_buf-> read.count       = privptr->p_env->readSize;

                        /*  initialize read_h command */
                        p_buf-> read_h.cmd_code = CCW_CLAW_CMD_READHEADER;
                        p_buf-> read_h.cda =
				(__u32)virt_to_phys(&(p_buf->header));
                        p_buf-> read_h.flags = CCW_FLAG_SLI | CCW_FLAG_CC;
                        p_buf-> read_h.count      = sizeof(CLAWH);

                        /*  initialize Signal command */
                        p_buf-> signal.cmd_code = CCW_CLAW_CMD_SIGNAL_SMOD;
                        p_buf-> signal.cda =
				(__u32)virt_to_phys(&(pClawH->flag));
                        p_buf-> signal.flags = CCW_FLAG_SLI | CCW_FLAG_CC;
                        p_buf-> signal.count     = 1;

                        /*  initialize RTIC1 command */
                        p_buf-> RTIC1.cmd_code = CCW_CLAW_CMD_TIC;
                        p_buf-> RTIC1.cda = (__u32)real_TIC_address;
                        p_buf-> RTIC1.flags = 0;
                        p_buf-> RTIC1.count      = 0;

                        /*  initialize r_read_ff command */
                        p_buf-> r_read_ff.cmd_code = CCW_CLAW_CMD_READFF;
                        p_buf-> r_read_ff.cda =
				(__u32)virt_to_phys(&(pClawH->flag));
                        p_buf-> r_read_ff.flags = 
				CCW_FLAG_SLI | CCW_FLAG_CC | CCW_FLAG_PCI;
                        p_buf-> r_read_ff.count    = 1;

                        /*    initialize RTIC2          */
                        memcpy(&p_buf->RTIC2, &p_buf->RTIC1, sizeof(ccw1_t));

                        /*     initialize Header     */
                        p_buf->header.length=0xffff;
                        p_buf->header.opcode=0xff;
                        p_buf->header.flag=CLAW_PENDING;
                
                        if (((unsigned long)p_buffer+privptr->p_env->readSize)
			    >= ((unsigned long)(p_buffer+2*
						(privptr->p_env->readSize)-1)
				& PAGE_MASK) ) {
			       	p_buffer= p_buffer+privptr->p_env->readSize;
                        }
                        else {
                                p_buffer=(void *)
					((unsigned long)(p_buffer+2*
							 (privptr->p_env->
							  readSize)-1) &
					 PAGE_MASK) ;
                        }
                }   /* for readBuffers   */
          }         /* readSize < PAGE_SIZE  */
          else {  /* read Size >= PAGE_SIZE  */
        
#ifdef DEBUGMSG
        printk(KERN_INFO "%s: init_ccw_bk() "
	       ">  Begin build claw read free chain \n",dev->name  );
#endif
                for (i=0 ; i< privptr->p_env->readBuffers ; i++) {
                        p_buffer = (void *)__get_free_pages(GFP_KERNEL|GFP_DMA,
				 privptr->p_buffer_pages_perread );
                        if (p_buffer==NULL) {
                                printk(KERN_INFO "claw: init_ccw_bk() "
				       "> __get_free_pages for read buf "
				       "failed : get is for %d pages\n", 
				       privptr->p_buffer_pages_perread );
                                free_pages((unsigned long)privptr->
					   p_buffer_ccw,
					   (int)privptr->p_buffer_ccw_num);
                                if (privptr->p_env->write_size < PAGE_SIZE) {
                                        free_pages((unsigned long)privptr->
						   p_buffer_write,
						   (int)privptr->
						   p_buffer_write_num);
                                }
                                else {
                                        p_buf=privptr->p_buffer_write;
                                        while (p_buf!=NULL) {
                                                free_pages((unsigned long)
							   p_buf->p_buffer,
							   (int)privptr->
						   p_buffer_pages_perwrite );
                                                p_buf=p_buf->next;
                                        }
                                }
                                privptr->p_buffer_ccw=NULL;
                                privptr->p_buffer_write=NULL;
#ifdef FUNCTRACE
                                printk(KERN_INFO "%s: init_ccw_bk() "
				       "> exit on line %d, rc = ENOMEM\n",
				       dev->name, __LINE__);
#endif                  
                                return -ENOMEM;
                        }
                        memset(p_buffer, 0x00, privptr->p_env->readSize);
                        p_buf        = p_free_chain;
                        privptr->p_buffer_read = p_buf;
                        p_free_chain = p_buf->next;

                        if (p_last_CCWB==NULL) {
                                p_buf->next=NULL;
                                real_TIC_address=0;
                                p_last_CCWB=p_buf;
                        }
                        else {
                                p_buf->next=p_first_CCWB;
                                real_TIC_address=
					(addr_t)virt_to_phys(&p_first_CCWB->
							     read );
                        }

                        p_first_CCWB=p_buf;

			// save virtual buffer address
                        p_buf->p_buffer=(PCLAWBUF)p_buffer;
                        /*  initialize read command */
                        p_buf-> read.cmd_code = CCW_CLAW_CMD_READ;
                        p_buf-> read.cda = (__u32)virt_to_phys(p_buffer);
                        p_buf-> read.flags = CCW_FLAG_SLI | CCW_FLAG_CC;
                        p_buf-> read.count       = privptr->p_env->readSize;

                        /*  initialize read_h command */
                        p_buf-> read_h.cmd_code = CCW_CLAW_CMD_READHEADER;
                        p_buf-> read_h.cda =
				(__u32)virt_to_phys(&(p_buf->header));
                        p_buf-> read_h.flags = CCW_FLAG_SLI | CCW_FLAG_CC;
                        p_buf-> read_h.count      = sizeof(CLAWH);     

                        /*  initialize Signal command */
                        p_buf-> signal.cmd_code = CCW_CLAW_CMD_SIGNAL_SMOD;
                        p_buf-> signal.cda =
				(__u32)virt_to_phys(&(pClawH->flag));
                        p_buf-> signal.flags = CCW_FLAG_SLI | CCW_FLAG_CC;
                        p_buf-> signal.count     = 1;

                        /*  initialize RTIC1 command */
                        p_buf-> RTIC1.cmd_code = CCW_CLAW_CMD_TIC;
                        p_buf-> RTIC1.cda = (__u32)real_TIC_address;
                        p_buf-> RTIC1.flags = 0;
                        p_buf-> RTIC1.count      = 0;
        
                        /*  initialize r_read_ff command */
                        p_buf-> r_read_ff.cmd_code = CCW_CLAW_CMD_READFF;
                        p_buf-> r_read_ff.cda =
				(__u32)virt_to_phys(&(pClawH->flag));
                        p_buf-> r_read_ff.flags =
				CCW_FLAG_SLI | CCW_FLAG_CC | CCW_FLAG_PCI;
                        p_buf-> r_read_ff.count    = 1;

                        /*    initialize RTIC2          */
                        memcpy(&p_buf->RTIC2, &p_buf->RTIC1, sizeof(ccw1_t));

                        /*     initialize Header     */
                        p_buf->header.length=0xffff;
                        p_buf->header.opcode=0xff;
                        p_buf->header.flag=CLAW_PENDING;
                
                }    /* For readBuffers   */
	  }     /*  readSize >= PAGE_SIZE   */
        }       /*  pBuffread = NULL */

        

        

#ifdef DEBUGMSG
        printk(KERN_INFO "%s: init_ccw_bk() "
	       ">  End build claw read free chain \n",dev->name  );
        p_buf=p_first_CCWB;
        while (p_buf!=NULL) {
                dumpit((char *)p_buf, sizeof(CCWBK));
                p_buf=p_buf->next;
        }
        
#endif

        add_claw_reads( dev  ,p_first_CCWB , p_last_CCWB);

#ifdef FUNCTRACE
        printk(KERN_INFO "%s: init_ccw_bk() "
	       "> exit on line %d\n",dev->name, __LINE__);
#endif
        return 0;
}    /*    end of init_ccw_bk */

/*-------------------------------------------------------------------*
*                                                                    *
*       probe_error                                                      *
*                                                                    *
*--------------------------------------------------------------------*/

static void probe_error( net_device *dev)
{
	PPRIV privptr;
	PENV  p_env;
	int    i;

#ifdef FUNCTRACE
        printk(KERN_INFO "%s: probe_error() > enter  \n",dev->name);
        printk(KERN_INFO "%s: variable dev =\n",dev->name);
        dumpit((char *) dev, sizeof(net_device));    
#endif

        privptr=dev->priv;
        p_env=privptr->p_env;

        for ( i=READ ; i<=WRITE ; i++) {
                FREE_IRQ(get_irq_by_devno(p_env->devno[i]),
			 privptr->channel[i].devstat);
                if (privptr->channel[i].devstat!=NULL) {
                        kfree(privptr->channel[READ].devstat);
                        privptr->channel[i].devstat=NULL;
                }
        }
        if (privptr->p_more_to_come_envelope!=NULL) {
                kfree(privptr->p_more_to_come_envelope);
                privptr->p_more_to_come_envelope=NULL;
        }
        if (privptr!=NULL) {
                kfree(privptr);
                privptr=NULL;
        }

#ifdef FUNCTRACE
        printk(KERN_INFO "%s: probe_error() > exit on line %d\n",
	       dev->name, __LINE__);
#endif

        return;
}    /*    probe_error    */



/*-------------------------------------------------------------------*
*    process_claw_control                                              *
*                                                                    *
*                                                                        *
*--------------------------------------------------------------------*/

static int process_claw_control( net_device *dev, PCCWBK p_ccw)
{

        PCLAWBUF p_buf;
        CLAWCTL  ctlbk;
        PCLAWCTL p_ctlbk;
        char    temp_host_name[8];
        char    temp_ws_name[8];
        PPRIV   privptr;
        PENV    p_env;
        PSYSVAL  p_sysval;
        PCONNECT p_connect=NULL;
        int rc;
        PCHBK   p_ch = NULL;

#ifdef FUNCTRACE
        printk(KERN_INFO "%s: process_claw_control() > enter  \n",dev->name);
        printk(KERN_INFO "%s: variable dev =\n",dev->name);
        dumpit((char *) dev, sizeof(net_device));
        printk(KERN_INFO "%s: variable p_ccw =\n",dev->name);
        dumpit((char *) p_ccw, sizeof(PCCWBK));   
#else
        udelay(1000); /* Wait a ms for the control packets to
			 catch up to each other */
#endif
        
        privptr=dev->priv;
        p_env=privptr->p_env;
        memcpy( &temp_host_name, p_env->host_name, 8);
        memcpy( &temp_ws_name, p_env->adapter_name, 8);

        printk(KERN_INFO "%s: CLAW device %.8s: Received Control Packet\n",
	       dev->name, temp_ws_name);
        if (privptr->RELEASE_PEND==1) {
#ifdef FUNCTRACE
                printk(KERN_INFO "%s: process_claw_control() "
		       "> exit on line %d, rc=0\n",dev->name, __LINE__);
#endif
                return 0;
        }
        p_buf=p_ccw->p_buffer;
        p_ctlbk=&ctlbk;
        memcpy(p_ctlbk, p_buf, sizeof(CLAWCTL));
#ifdef IOTRACE
        printk(KERN_INFO "%s: dump claw control data \n",dev->name);
        dumpit((char *)p_ctlbk, sizeof(CLAWCTL));
#endif

        switch (p_ctlbk->command)
        {
                case SYSTEM_VALIDATE_REQUEST:
                        printk( "%s: Sys Validate Request: Vers=%d,"
				"link_id=%d,Corr=%d,WS name=%.8s,"
				"Host name=%.8s\n",
                                dev->name, p_ctlbk->version,
				p_ctlbk->linkid, p_ctlbk->correlator,
				temp_ws_name,temp_host_name);
                        if (p_ctlbk->version!=CLAW_VERSION_ID) {
				send_system_validate_response(dev, p_ctlbk,
						      CLAW_RC_WRONG_VERSION );
                                printk("%s:  %d is wrong version id. "
				       "Expected %d\n",
				       dev->name, p_ctlbk->version,
				       CLAW_VERSION_ID);
                        }
                        p_sysval=(PSYSVAL)&(p_ctlbk->data);

                        if (0!=memcmp(temp_host_name,p_sysval->Host_name,8 )) {
                                send_system_validate_response(dev, p_ctlbk,
						      CLAW_RC_NAME_MISMATCH );
                                printk(KERN_INFO "%s:  Host name mismatch\n",
				       dev->name);
                        }
                        if (0!=memcmp(temp_ws_name,p_sysval->WS_name,8)) {
                                send_system_validate_response(dev, p_ctlbk,
						      CLAW_RC_NAME_MISMATCH );
                                printk(KERN_INFO "%s: WS name mismatch\n",
				       dev->name);
                        }
                        if ( p_sysval->Write_frame_size < p_env->write_size) {
                                send_system_validate_response(dev, p_ctlbk,
					      CLAW_RC_HOST_RCV_TOO_SMALL );
                                printk(KERN_INFO "%s: host write size is "
				       "too small\n", dev->name);
                        }
                        if ( p_sysval->read_frame_size < p_env->readSize) {
                                send_system_validate_response(dev, p_ctlbk,
					      CLAW_RC_HOST_RCV_TOO_SMALL );
                                printk(KERN_INFO "%s: host read size is "
				       "too small\n", dev->name);
                        }

                        send_system_validate_response(dev, p_ctlbk, 0 );
                        printk("%s: CLAW device %.8s: System validate "
			       "completed.\n", dev->name, temp_ws_name);
                        privptr->SYSTEM_VALIDATE_COMP=1;
                
                        start_connect_request(dev);
                        break;
                        
                case SYSTEM_VALIDATE_RESPONSE:
                        printk("%s: Sys Validate Resp: Vers=%d,Corr=%d,"
			       "RC=%d,WS name=%.8s,Host name=%.8s\n",
                        dev->name,
                        p_ctlbk->version,
                        p_ctlbk->correlator,
                        p_ctlbk->rc,
                        temp_ws_name,
                        temp_host_name);

                        switch (p_ctlbk->rc)
                        {
                                case 0:
                                        privptr->SYSTEM_VALIDATE_COMP=1;
                                        printk(KERN_INFO "%s: CLAW device "
					       "%.8s: System validate "
					       "completed.\n",
                                                dev->name, temp_ws_name);
                                        start_connect_request(dev);
                                        break;
                                case CLAW_RC_NAME_MISMATCH:
                                        printk(KERN_INFO "%s: Sys Validate "
					       "Resp :  Host, WS name is "
					       "mismatch\n",
                                                dev->name);
                                        break;
                                case CLAW_RC_WRONG_VERSION:
                                        printk(KERN_INFO "%s: Sys Validate "
					       "Resp :  Wrong version\n",
					       dev->name);
                                        break;
                                case CLAW_RC_HOST_RCV_TOO_SMALL:
                                        printk(KERN_INFO "%s: Sys Validate "
					       "Resp :  bad frame size\n",
					       dev->name);
                                        break;
                                default:
                                        printk(KERN_INFO "%s: Sys Validate "
					       "error code=%d \n",
					       dev->name, p_ctlbk->rc );
                                        break;
                        }
                        break;
                        
                case CONNECTION_REQUEST:
                        p_connect=(PCONNECT)&(p_ctlbk->data);
                        printk(KERN_INFO "%s: Conn Req: Vers=%d,link_id=%d,"
			       "Corr=%d,HOST appl=%.8s,WS appl=%.8s\n",
                        dev->name,
                        p_ctlbk->version,
                        p_ctlbk->linkid,
                        p_ctlbk->correlator,
                        p_connect->Host_name,
                        p_connect->WS_name);
                        if (privptr->ActiveLinkID!=0 ) {
                                send_disconnect(dev, p_ctlbk);
                                printk(KERN_INFO "%s: Conn Req error : "
				       "already logical link is active \n",
				       dev->name);
                        }
                        if (p_ctlbk->linkid!=1 ) {
                                send_disconnect(dev, p_ctlbk);
                                printk(KERN_INFO "%s: Conn Req error : "
				       "requesting logical link id is "
				       "not 1\n", dev->name);
                        }

                        rc=find_link(dev, p_connect->Host_name,
				     p_connect->WS_name);
                        if (rc!=0) {
                                send_disconnect(dev, p_ctlbk);
                                printk(KERN_INFO "%s: Conn Req error : "
				       "requesting appl name does not match\n",
				       dev->name);
                        }

                        send_claw_control(dev, CONNECTION_CONFIRM,
					  p_ctlbk->linkid,p_ctlbk->correlator,
					  0, p_connect->Host_name,
					  p_connect->WS_name);

                        printk(KERN_INFO "%s: CLAW device %.8s: "
			       "Connection completed link_id=%d.\n",
			       dev->name, temp_ws_name,
			       p_ctlbk->linkid);
                        privptr->ActiveLinkID=p_ctlbk->linkid;
                        p_ch=&privptr->channel[WRITE];
                        wake_up(&p_ch->wait);  /* wake up claw_open ( WRITE) */
                        break;
                case CONNECTION_RESPONSE:
                        p_connect=(PCONNECT)&(p_ctlbk->data);
                        printk(KERN_INFO "%s: Conn Resp: Vers=%d,link_id=%d,"
			       "Corr=%d,RC=%d,Host appl=%.8s, WS appl=%.8s\n",
			       dev->name, p_ctlbk->version, p_ctlbk->linkid,
			       p_ctlbk->correlator, p_ctlbk->rc,
			       p_connect->Host_name,p_connect->WS_name);

                        if (p_ctlbk->rc !=0 ) {
                                printk(KERN_INFO "%s: Conn Resp error: "
				       "rc=%d \n", dev->name, p_ctlbk->rc);
                                return 1;
                        }
                        rc=find_link(dev, p_connect->Host_name,
				     p_connect->WS_name);
                        if (rc!=0) {
                                send_disconnect(dev, p_ctlbk);
                                printk(KERN_INFO "%s:  Conn Resp error: "
				       "requesting appl name does not match\n",
				       dev->name);
                        }
			// should be negative until CONNECTION_CONFIRM
                        privptr->ActiveLinkID =  - (p_ctlbk->linkid);  
                        break;
                case CONNECTION_CONFIRM:
                        p_connect=(PCONNECT)&(p_ctlbk->data);
                        printk(KERN_INFO "%s: Conn Confirm: Vers=%d,"
			       "link_id=%d,Corr=%d,Host appl=%.8s,"
			       "WS appl=%.8s\n",
			       dev->name,
			       p_ctlbk->version,
			       p_ctlbk->linkid,
			       p_ctlbk->correlator,
			       p_connect->Host_name,
			       p_connect->WS_name);

                        if (p_ctlbk->linkid== -(privptr->ActiveLinkID)) {
                                privptr->ActiveLinkID=p_ctlbk->linkid;
                                p_ch=&privptr->channel[WRITE];
                                wake_up(&p_ch->wait);  /* wake up
							  claw_open ( WRITE) */
                        }
                        else { 
                                printk(KERN_INFO "%s: Conn confirm: "
				       "unexpected linkid=%d \n",
				       dev->name, p_ctlbk->linkid);
                                send_disconnect(dev, p_ctlbk);
                        }
                        break;
                case DISCONNECT:
                        printk(KERN_INFO "%s: Disconnect: Vers=%d,link_id=%d,"
			       "Corr=%d\n", dev->name, p_ctlbk->version,
			       p_ctlbk->linkid, p_ctlbk->correlator);
                        privptr->ActiveLinkID=0;
                        break;
                case CLAW_ERROR:
                        printk(KERN_INFO "%s:  CLAW ERROR detected\n",
			       dev->name );
                        break;
                default:
                        printk(KERN_INFO "%s:  Unexpected command code=%d\n",
			       dev->name,  p_ctlbk->command);
                        break;
        }

#ifdef FUNCTRACE
        printk(KERN_INFO "%s: process_claw_control() "
	       "> exit on line %d, rc = 0\n",dev->name, __LINE__);
#endif

        return 0;
}   /*    end of process_claw_control    */


/*-------------------------------------------------------------------*
*               send_claw_control                                      *
*                                                                    *
*--------------------------------------------------------------------*/

static int send_claw_control(net_device *dev, __u8 type, __u8 link,
			     __u8 correlator, __u8 rc, char *local_name,
			     char *remote_name)
{       
        PPRIV                           privptr;
        PCLAWCTL                        p_ctl;
        PSYSVAL                         p_sysval;
        PCONNECT                        p_connect;
        struct sk_buff *skb;


#ifdef FUNCTRACE
        printk(KERN_INFO "%s: send_claw_control() > enter  \n",dev->name);
        printk(KERN_INFO "%s: variable type = %X, link = %X, "
	       "correlator = %X, rc = %x\n",
                dev->name,type, link, correlator, rc);
        printk(KERN_INFO "%s: variable local_name = %s, remote_name = %s\n",
	       dev->name, local_name, remote_name);
#endif
        privptr=dev->priv;
        p_ctl=(PCLAWCTL)&privptr->ctl_bk;

        p_ctl->command=type;
        p_ctl->version=CLAW_VERSION_ID;
        p_ctl->linkid=link;
        p_ctl->correlator=correlator;
        p_ctl->rc=rc;

        p_sysval=(PSYSVAL)&p_ctl->data;
        p_connect=(PCONNECT)&p_ctl->data;

        switch (p_ctl->command) {
                case SYSTEM_VALIDATE_REQUEST:
                case SYSTEM_VALIDATE_RESPONSE:
                        memcpy(&p_sysval->Host_name, local_name, 8);
                        memcpy(&p_sysval->WS_name, remote_name, 8);
                        p_sysval->read_frame_size=privptr->p_env->readSize;
                        p_sysval->Write_frame_size=privptr->p_env->write_size;
                        memset(&p_sysval->reserved, 0x00, 4);
                        break;
                case CONNECTION_REQUEST:
                case CONNECTION_RESPONSE:
                case CONNECTION_CONFIRM:
                case DISCONNECT:                                
                        memcpy(&p_sysval->Host_name, local_name, 8);
                        memcpy(&p_sysval->WS_name, remote_name, 8);

                        memset(&p_connect->reserved1, 0x00, 4);
                        memset(&p_connect->reserved2, 0x00, 4);
                        break;
                default:
                        break;
        }
                
        /*      write Control Record to the device             */
                                

        skb = dev_alloc_skb(sizeof(CLAWCTL)); 
        if (!skb) {
                printk(  "%s: send_claw_control() low on mem, "
			 "returning...\n",dev->name);
#ifdef DEBUG
                printk(KERN_INFO "%s: send_claw_control() "
		       "> exit, rc = ENOMEM\n",dev->name);
#endif
                return -ENOMEM;
        }

        memcpy(skb_put(skb, sizeof(CLAWCTL)), p_ctl, sizeof(CLAWCTL));
#ifdef IOTRACE
        dumpit((char *)p_ctl,sizeof(CLAWCTL));
#endif

        claw_hw_tx(skb, dev, 0);

#ifdef FUNCTRACE
        printk(KERN_INFO "%s: send_claw_control() > exit on line %d\n",
	       dev->name, __LINE__);
#endif

        return 0;
}  /*   end of send_claw_control  */

/*-------------------------------------------------------------------*
*               send_connect_request                                   *
*                                                                    *
*--------------------------------------------------------------------*/
static int send_connect_request(net_device *dev, __u8 link)
{
        int                rc;
        PPRIV           privptr=dev->priv;
        PCLAWCTL           p_ctl;

#ifdef FUNCTRACE
        printk(KERN_INFO "%s: send_connect_request() > enter  \n",dev->name);
        printk(KERN_INFO "%s: variable link = %X, dev =\n",dev->name, link);
        dumpit((char *) dev, sizeof(net_device));
#endif

        p_ctl=(PCLAWCTL)&privptr->ctl_bk;

        if ( privptr-> SYSTEM_VALIDATE_COMP==0x00 ) {
#ifdef FUNCTRACE
                printk(KERN_INFO "%s: send_claw_control() "
		       "> exit on line %d, rc = 1\n",dev->name, __LINE__);
#endif   
                return 1;
        }
        rc=send_claw_control(dev, CONNECTION_REQUEST,0,0,0,
                           HOST_APPL_NAME, privptr->p_env->ApiType);

#ifdef FUNCTRACE
        printk(KERN_INFO "%s: send_connect_request() "
	       "> exit on line %d, rc = %d\n",dev->name, __LINE__, rc);
#endif
        return rc;

}  /*  end of send_connect_request */


/*-------------------------------------------------------------------*
*               send_disconnect                                       *
*                                                                    *
*--------------------------------------------------------------------*/

static int send_disconnect(net_device *dev, PCLAWCTL p_ctl)
{
        int rc;
        PCONNECT  p_connect;

#ifdef FUNCTRACE
        printk(KERN_INFO "%s: send_disconnect() > enter\n",dev->name);
        printk(KERN_INFO "%s: variable dev =\n",dev->name);
        dumpit((char *) dev, sizeof(net_device));
        printk(KERN_INFO "%s: variable p_ctl",dev->name);
        dumpit((char *) p_ctl, sizeof(CLAWCTL));
#endif
        p_connect=(PCONNECT)&p_ctl->data;

        rc=send_claw_control(dev, DISCONNECT, p_ctl->linkid, 
			     p_ctl->correlator, 0, 
			     p_connect->Host_name, p_connect->WS_name);
#ifdef FUNCTRACE
        printk(KERN_INFO "%s: send_disconnect() "
	       "> exit on line %d, rc = %d\n",dev->name, __LINE__, rc);
#endif
        return rc;
}     /*   end of send_disconnect    */


/*-------------------------------------------------------------------*
*               send_system_validate_response                           *
*                                                                    *
*--------------------------------------------------------------------*/

static int send_system_validate_response(net_device *dev, PCLAWCTL p_ctl,
					 __u32 return_code)
{
        PENV  p_env;
        PPRIV  privptr;
        int    rc;

#ifdef FUNCTRACE
        printk(KERN_INFO "%s: send_system_validate_response() "
	       "> enter  \n",dev->name);
        printk(KERN_INFO "%s: variable return_code = %d, dev =\n",
	       dev->name, return_code);
        dumpit((char *) dev, sizeof(net_device));
        printk(KERN_INFO "%s: variable p_ctl =\n",dev->name);
        dumpit((char *) p_ctl, sizeof(CLAWCTL));
#endif
        privptr = dev->priv;
        p_env=privptr->p_env;
        rc=send_claw_control(dev, SYSTEM_VALIDATE_RESPONSE,
			     p_ctl->linkid, p_ctl->correlator,
			     return_code, p_env->host_name,
			     p_env->adapter_name );
#ifdef FUNCTRACE
        printk(KERN_INFO "%s: send_system_validate_response() "
	       "> exit on line %d, rc = %d\n",dev->name, __LINE__, rc);
#endif
        return rc;
}     /*    end of send_system_validate_response    */

/*-------------------------------------------------------------------*
*               start_connect_request                                  *
*                                                                    *
*--------------------------------------------------------------------*/

static int start_connect_request(net_device *dev )
{
        int rc;
  
#ifdef FUNCTRACE
        printk(KERN_INFO "%s: send_connect_request() > enter  \n",dev->name);
        printk(KERN_INFO "%s: variable dev =\n",dev->name);
        dumpit((char *) dev, sizeof(net_device));
#endif

        rc=send_connect_request(dev, 1);

#ifdef FUNCTRACE
        printk(KERN_INFO "%s: send_connect_request() "
	       "> exit on line %d, rc = %d\n",dev->name, __LINE__, rc);
#endif          
        return rc;
}    /*   end of start_connect_request   */



/*-------------------------------------------------------------------*
 *   claw_stats                                                      *
 *-------------------------------------------------------------------*/

struct net_device_stats *claw_stats(net_device *dev)
{
        PPRIV   privptr;
#ifdef FUNCTRACE
        printk(KERN_INFO "%s: claw_stats() > enter\n",dev->name);
        printk(KERN_INFO "%s: variable dev =\n",dev->name);
        dumpit((char *) dev, sizeof(net_device));
#endif           
        privptr = dev->priv;
#ifdef FUNCTRACE
        printk(KERN_INFO "%s: claw_stats() > exit on line %d\n",
	       dev->name, __LINE__);
#endif
        return &privptr->stats;
}     /*   end of claw_stats   */


/*-------------------------------------------------------------------*
*       unpack_read                                                  *
*                                                                    *
*--------------------------------------------------------------------*/
static void unpack_read (net_device *dev )
{
        struct sk_buff *skb;
        PPRIV   privptr;
        PCCWBK  p_this_ccw;
        PCCWBK  p_first_CCW;
        PCCWBK  p_last_CCW;
        __u32   length_of_data;
        __u32   link_number;
        __u8    mtc_this_frame=0;
        __u32   bytes_to_move;
        PCHBK   p_ch = NULL;
        int             i=0;

#ifdef FUNCTRACE
        printk(KERN_INFO "%s: unpack_read() > enter  \n",dev->name);
        printk(KERN_INFO "%s: variable dev =\n",dev->name);
        dumpit((char *) dev, sizeof(net_device));
#endif

        p_first_CCW=NULL;
        p_last_CCW=NULL;
        privptr=dev->priv;
        p_this_ccw=privptr->p_read_active_first;
        i=0;
        while   (p_this_ccw!=NULL && p_this_ccw->header.flag!=CLAW_PENDING) {
                /*      Update read statistics   */
#ifdef IOTRACE
                printk(KERN_INFO "claw: dumping p_this_ccw \n");
                dumpit((char*)p_this_ccw, sizeof(CCWBK));
                printk(KERN_INFO "claw: dumping p_this_ccw->p_buffer(32) \n");
                dumpit((char *)p_this_ccw->p_buffer, 32 );
#endif
                link_number=p_this_ccw->header.opcode / 8;
                if ((p_this_ccw->header.opcode & MORE_to_COME_FLAG)!=0) {
#ifdef DEBUGMSG
                        printk(KERN_INFO "%s: unpack_read "
			       "> More_to_come is ON  \n",dev->name);
#endif
                        mtc_this_frame=1;
                        if (p_this_ccw->header.length!=
			    privptr->p_env->readSize ) {
                                printk(KERN_INFO " %s: Invalid frame "
				       "detected length is %02x\n" ,
				       dev->name, p_this_ccw->header.length);
                        }
                }
                

                if (privptr->MoreToComeSkipping) {
                        /*                      
                        * We're in the mode of skipping past a multi-frame 
			* message that we can't process for some reason
			* or other.  The first frame without the
			* More-To-Come flag is the last frame of the
			* skipped message.
                        */
                        /* in case of More-To-Come not set in this frame */ 
                        if (mtc_this_frame==0) {
				// Ok, that's the end
                                privptr->MoreToComeSkipping=0;  
                                privptr->more_to_come_logical_link=-1; 
                        }
#ifdef DEBUGMSG
                        printk(KERN_INFO "%s: unpack_read "
			       "> goto next frame from MoretoComeSkip \n",
			       dev->name);
#endif
                        goto NextFrame;
                }
                        
                if (link_number==0) {
                        process_claw_control(dev, p_this_ccw);
#ifdef DEBUGMSG
                        printk(KERN_INFO "%s: unpack_read "
			       "> goto next frame from "
			       "process_claw_control \n",dev->name);
#endif
                        goto NextFrame;
                }

                bytes_to_move=p_this_ccw->header.length;

                if (privptr->more_to_come_logical_link<0) {
#ifdef DEBUGMSG
		   	printk(KERN_INFO "%s: unpack_read "
			       "> more_to_come_logical_link < 0  \n",dev->name);
#endif

			/*                              
                           if More-To-Come is set in this frame then we don't 
    		           know length of entire message, and hence have to 
    		           allocate large buffer
			 */
                
			/*      We are starting a new envelope  */
			privptr->more_to_come_offset=0;
			privptr->more_to_come_logical_link=link_number;
	    	}

                if (bytes_to_move > (MAX_ENVELOPE_SIZE-
				   privptr->more_to_come_offset) ) {
                        /*      error     */
#ifdef DEBUGMSG
                        printk(KERN_INFO "%s: unpack_read "
			       "> goto next frame from MoretoComeSkip \n",
			       dev->name);
                        printk(KERN_INFO "      bytes_to_move "
			       "> (MAX_ENVELOPE_SIZE-"
			       "privptr->more_to_come_offset)\n");
#endif
                        privptr->stats.rx_frame_errors++;
                        goto NextFrame;
                }
                memcpy( privptr->p_more_to_come_envelope+
			privptr->more_to_come_offset, 
                        p_this_ccw->p_buffer, bytes_to_move);
#ifdef DEBUGMSG
                printk(KERN_INFO "%s: unpack_read > moved data \n",dev->name);
                dumpit((char *)p_this_ccw->p_buffer, 64 );
		printk(KERN_INFO "%s: unpack_read > bytelength %d \n",
		       dev->name,bytes_to_move);
#endif
                if (mtc_this_frame==0) {
                        length_of_data=privptr->more_to_come_offset+
				bytes_to_move;
                        skb=dev_alloc_skb(length_of_data);
                        if (skb) {
                                memcpy(skb_put(skb,length_of_data),
				       privptr->p_more_to_come_envelope,
				       length_of_data);
                                skb->mac.raw=skb->data;
                                skb->dev=dev;
                                skb->protocol=htons(ETH_P_IP);
                                skb->ip_summed=
#ifdef CLAW_CHANDEV
                                        ( privptr->checksum_received_ip_pkts 
                                          ? CHECKSUM_NONE : 
					  CHECKSUM_UNNECESSARY );
#else
                                CHECKSUM_UNNECESSARY;
#endif
                                privptr->stats.rx_packets++;
                                netif_rx(skb);
#ifdef DEBUGMSG
                                printk(KERN_INFO "%s: unpack_read "
				       "> netif_rx(skb) completed \n",
				       dev->name);
#endif
                        }
                        else {
                                privptr->stats.rx_dropped++;
                                printk(KERN_WARNING "%s: unpack_read() "
				       "low on memory\n",dev->name);
                        }
                        privptr->more_to_come_offset=0;
                        privptr->more_to_come_logical_link=-1;
                } 
                else {
                        privptr->more_to_come_offset+=bytes_to_move;
                }

NextFrame:

                /*
                *       Remove ThisCCWblock from active read queue, and add it
                *       to queue of free blocks to be reused.
                *
                */
                i++;
		// remove this from queue
                privptr->p_read_active_first=p_this_ccw->next;    
                p_this_ccw->next=NULL;        // terminate chain
		// show no longer completed
                p_this_ccw->header.flag=CLAW_PENDING; 
                p_this_ccw->header.length=0xffff;
                p_this_ccw->header.opcode=0xff;
                /*
                *       add this one to the free queue for later reuse
                */
                if (p_first_CCW==NULL) {
                        p_first_CCW=p_this_ccw;
                }
                else {
                        p_last_CCW->next=p_this_ccw;
                }
                p_last_CCW=p_this_ccw;
                /*
                *       chain to next block on active read queue
                */
                p_this_ccw=privptr->p_read_active_first;
        } /* end of while */

        /*      check validity                  */

#ifdef DEBUGMSG
        printk(KERN_INFO "%s: unpack_read > processed frame is %d \n",
	       dev->name,i);
        printk(KERN_INFO "%s: unpack_read > F:%x L:%x\n",
	       dev->name,(unsigned long)p_first_CCW,(unsigned long)p_last_CCW);
#endif
        add_claw_reads(dev, p_first_CCW, p_last_CCW);
        
        p_ch=&privptr->channel[READ];
        start_read(dev, LOCK_YES);

#ifdef FUNCTRACE
        printk(KERN_INFO "%s: unpack_read() "
	       "> exit on line %d\n",dev->name, __LINE__);
#endif
        return;
}     /*  end of unpack_read   */

/*-------------------------------------------------------------------*
*       start_read                                                    *
*                                                                    *
*--------------------------------------------------------------------*/
static void start_read ( net_device *dev, int lock )
{
        int        rc = 0;
        __u32      parm;
        __u8       flags = DOIO_REPORT_ALL | DOIO_DENY_PREFETCH ;
        unsigned long  saveflags;
        PPRIV      privptr=(PPRIV) dev->priv;             
        PCCWBK     p_ccwbk;
        PCHBK      p_ch;
        PCLAWH     p_clawh;

        p_ch=&privptr->channel[READ];
        
#ifdef FUNCTRACE
        printk(KERN_INFO "%s: start_read() > enter  \n",dev->name);
        printk(KERN_INFO "%s: variable lock = %d, dev =\n",dev->name, lock);
        dumpit((char *) dev, sizeof(net_device));
#endif

        p_clawh=(PCLAWH)privptr->pClawSignalBlock;
        p_clawh->flag=CLAW_IDLE;    /* 0x00 */

        if ((privptr->p_write_active_first!=NULL &&
             privptr->p_write_active_first->header.flag!=CLAW_PENDING) ||
            (privptr->p_read_active_first!=NULL &&
             privptr->p_read_active_first->header.flag!=CLAW_PENDING )) {
                p_clawh->flag=CLAW_BUSY;    /* 0x00 */
        } 
        else {
                p_clawh->flag=CLAW_IDLE;    /* 0x00 */
        }
           


#ifdef DEBUGMSG
        printk(KERN_INFO "%s: start_read - state-%02x\n",
	       dev->name, p_ch->claw_state);
#endif

        if (lock==LOCK_YES) {
                s390irq_spin_lock_irqsave(p_ch->irq, saveflags);
        } 


        if (test_and_set_bit(0, (void *)&p_ch->IO_active) == 0) {

#ifdef DEBUGMSG
                printk(KERN_INFO "%s: HOT READ started in "
		       "start_read routine\n" ,dev->name);
                p_clawh=(PCLAWH)privptr->pClawSignalBlock;
                dumpit((char *)&p_clawh->flag , 1);             
#endif 
                p_ccwbk=privptr->p_read_active_first;
                parm = (unsigned long) p_ch;
                rc = do_IO (p_ch->irq, &p_ccwbk->read, parm, 0xff, flags );
                if (rc != 0) { 
                        ccw_check_return_code(dev, rc);
                }
        }

        if (lock==LOCK_YES) {
                s390irq_spin_unlock_irqrestore(p_ch->irq, saveflags);
        }
 
#ifdef FUNCTRACE
        printk(KERN_INFO "%s: start_read > exit on line %d\n",
	       dev->name, __LINE__);
#endif
        return;
}       /*    end of start_read    */

/*-------------------------------------------------------------------*
*       start_output_io                                                *
*                                                                    *
*--------------------------------------------------------------------*/

static void start_output_io( net_device *dev )
{
        int             rc = 0;
        unsigned long   parm;
        __u8            flags = DOIO_DENY_PREFETCH;
        PPRIV           privptr=(PPRIV) dev->priv;             
        PCHBK           p_ch;
        PCCWBK          p_first_CCW;

#ifdef FUNCTRACE
        printk(KERN_INFO "%s: start_output_io() > enter  \n",dev->name );
        printk(KERN_INFO "%s: variable dev =\n",dev->name);
        dumpit((char *) dev, sizeof(net_device));
#endif
        privptr=dev->priv;   
        p_ch=&privptr->channel[WRITE];

#ifdef DEBUGMSG
        printk(KERN_INFO "%s: start_output_io - state-%02x\n" ,
	       dev->name, p_ch->claw_state);
#endif 
        
        p_first_CCW=privptr->p_write_active_first;

        if (p_ch->claw_state == CLAW_STOP)
                return;
        if (p_first_CCW == NULL) {
#ifdef FUNCTRACE
                printk(KERN_INFO "%s: start_output_io() > exit on line %d\n",
		       dev->name, __LINE__);
#endif                  
                return;
        }


        if (test_and_set_bit(0, (void *)&p_ch->IO_active) == 0) {
                parm = (unsigned long) p_ch;
#ifdef DEBUGMSG
                printk(KERN_INFO "%s: start_output_io - do_io \n" ,dev->name);
                dumpit((char *)p_first_CCW, sizeof(CCWBK));
#endif           
                rc = do_IO (p_ch->irq, &p_first_CCW->Write, parm,0xff,flags);
                if (rc != 0) {  
                        ccw_check_return_code(dev, rc);
                }
        }

        dev->trans_start = jiffies;
#ifdef FUNCTRACE
        printk(KERN_INFO "%s: start_output_io() "
	       "> exit on line %d\n",dev->name, __LINE__);
#endif

        return;
}       /*    end of start_output_io    */

/*-------------------------------------------------------------------*
*       Free write buffers                                           *
*                                                                    *
*--------------------------------------------------------------------*/

static void free_write_buffers( net_device *dev )
{

        PPRIV      privptr=(PPRIV) dev->priv;             
        PCCWBK     p_first_CCW,p_last_CCW,p_this_ccw,pNextCCW;
#ifdef IOTRACE
        PCCWBK          p_buf;
#endif  
#ifdef FUNCTRACE
        printk(KERN_INFO "%s: free_write_buffers() > enter\n",dev->name);
        printk(KERN_INFO "%s: free count = %d  variable dev =\n",
	       dev->name,privptr->write_free_count);
        dumpit((char *) dev, sizeof(net_device));
#endif

        /*  scan the write queue to free any completed write packets   */
        p_first_CCW=NULL;
        p_last_CCW=NULL;
                

#ifdef IOTRACE
        printk(KERN_INFO "%s:  Dump current CCW chain \n",dev->name  );
        p_buf=privptr->p_write_active_first;
        while (p_buf!=NULL) {
                dumpit((char *)p_buf, sizeof(CCWBK));
                p_buf=p_buf->next;
        }
        if (p_buf==NULL) {
                printk(KERN_INFO "%s: privptr->p_write_active_first==NULL\n",
		       dev->name  );
        }
        p_buf=(PCCWBK)privptr->p_end_ccw;
        dumpit((char *)p_buf, sizeof(ENDCCW));
#endif
        p_this_ccw=privptr->p_write_active_first;
        while ( (p_this_ccw!=NULL) && (p_this_ccw->header.flag!=CLAW_PENDING))
        {
                pNextCCW = p_this_ccw->next; // Look ahead 1 block      @A1A
                if (((pNextCCW!=NULL) && 
		     (pNextCCW->header.flag!=CLAW_PENDING)) ||                  
                    ((p_this_ccw == privptr->p_write_active_last) && 
                     (p_this_ccw->header.flag!=CLAW_PENDING))) {    /*  @A1A*/
                        /* The next CCW is OK or this is the last CCW...
			 * free it                            @A1A  */ 
			// Remove this from queue
                        privptr->p_write_active_first=p_this_ccw->next;
			// show no longer completed
                        p_this_ccw->header.flag=CLAW_PENDING;
			//      pick up free chain
                        p_this_ccw->next=privptr->p_write_free_chain;
			//      add to free chain
                        privptr->p_write_free_chain=p_this_ccw;
			//      add 1
                        privptr->write_free_count++;
			// check new first entry
                        p_this_ccw=privptr->p_write_active_first;
                        privptr->stats.tx_packets++;
                }       /* @A1A  */
                else {  /* @A1A  */                             
                        break;    /* @A1A  */
                }
        }
        if (privptr->write_free_count!=0) {
                claw_clearbit_busy(TB_NOBUFFER,dev);
        }

        /*   whole chain removed   */
        if (privptr->p_write_active_first==NULL) {
                privptr->p_write_active_last=NULL;
#ifdef DEBUGMSG
                printk(KERN_INFO "%s: free_write_buffers> "
		       "p_write_active_first==NULL\n",dev->name);
#endif
        }



#ifdef IOTRACE
        printk(KERN_INFO "%s: Dump arranged CCW chain \n",dev->name  );
        p_buf=privptr->p_write_active_first;
        while (p_buf!=NULL) {
                dumpit((char *)p_buf, sizeof(CCWBK));
                p_buf=p_buf->next;
        }
        if (p_buf==NULL) {
                printk(KERN_INFO "%s: privptr->p_write_active_first==NULL\n",
		       dev->name  );
        }
        p_buf=(PCCWBK)privptr->p_end_ccw;
        dumpit((char *)p_buf, sizeof(ENDCCW));
#endif


#ifdef FUNCTRACE
        printk(KERN_INFO "%s: free_write_buffers > free count = %d\n",
	       dev->name,privptr->write_free_count);
        printk(KERN_INFO "%s: free_write_buffers() > exit on line %d\n",
	       dev->name, __LINE__);
#endif
        return;
}


#ifdef CLAW_CHANDEV
static void claw_msck_notify(void *dev, int msck_irq,
			     chandev_msck_status prevstatus, 
			     chandev_msck_status newstatus)
{

}
#endif


#ifndef CLAW_CHANDEV
/*--------------------------------------------------------------------*
*    claw_init_dev                                                    *
*---------------------------------------------------------------------*/
static __init int claw_init_dev(char *parm)
{
        int n_parm, rc = 0;
        PENV p_env = 0;
        net_device *dev;
        unsigned long dev_num = 0, dev_addr = 0, num_rbuf = 0, num_wbuf = 0;
        char *hostname = NULL, *adaptername = NULL, *apitype = NULL;
        char *next, *endp;
        

#ifdef FUNCTRACE
        printk(KERN_INFO "claw: claw_init_dev() > enter \n");
        printk(KERN_INFO "claw: variable parm =\n");
        dumpit(parm, 256);
#endif

        /* Parse parameter string */
        n_parm = 0;
        next = parm;
        while ( (parm = next) != NULL )
        {
                next = strchr(parm, ':');
                if (next)
                        *next++ = 0;

                switch (n_parm++)
                {
                case 0:  /* device number */
                        dev_num = simple_strtoul(parm, &endp, 0);
                        if ( *endp ) {
                                printk(KERN_WARNING "claw: dev id not "
				       "numeric\n");
                                rc = -EINVAL;
                                goto error;
                        }
                        if (dev_num >= CLAW_MAX_DEV) {
                                printk(KERN_WARNING "claw: dev id "
				       "exceeds %d\n", CLAW_MAX_DEV-1);
                                rc = -EINVAL;
                                goto error;
                        }
                        break;

                case 1:  /* device address */
                        dev_addr = simple_strtoul(parm, &endp, 0);
                        if ( *endp ) {
                                printk(KERN_WARNING "claw: dev address "
				       "not numeric\n");
                                rc = -EINVAL;
                                goto error;
                        }
                        break;

                case 2:  /* number read buffers */
                        num_rbuf = simple_strtoul(parm, &endp, 0);
                        if ( *endp || num_rbuf == 0 ) {
                                printk(KERN_WARNING "claw: invalid nr "
				       "read buffers\n");
                                rc = -EINVAL;
                                goto error;
                        }
                        break;

                case 3:  /* number write buffers */
                        num_wbuf = simple_strtoul(parm, &endp, 0);
                        if ( *endp || num_wbuf == 0 ) {
                                printk(KERN_WARNING "claw: invalid nr "
				       "write buffers\n");
                                rc = -EINVAL;
                                goto error;
                        }
                        break;

                case 4:  /* host name */
                        hostname = parm;
                        if ( strlen(hostname) > MAX_NAME_LEN ) {
                                printk(KERN_WARNING "claw: host "
				       "name too long\n");
                                rc = -EINVAL;
                                goto error;
                        }
                        break;

                case 5:  /* adapter name */
                        adaptername = parm;
                        if ( strlen(adaptername) > MAX_NAME_LEN ) {
                                printk(KERN_WARNING "claw: adapter "
				       "name too long\n");
                                rc = -EINVAL;
                                goto error;
                        }
                        break;

                case 6:  /* api type */
                        apitype = parm;
                        if ( strlen(apitype) > MAX_NAME_LEN ) {
                                printk(KERN_WARNING "claw: api type "
				       "too long\n");
                                rc = -EINVAL;
                                goto error;
                        }
                        break;

                default:
                        printk(KERN_WARNING "claw: wrong number of "
			       "parameters passed\n");
                        rc = -EINVAL;
                        goto error;
                }
        }
        
        /* Check read device address */
        rc = claw_channel_check(dev_addr);
        if (rc != 0) 
        {
                printk(KERN_WARNING "claw%ld: Can not use CLAW device "
		       "address of %lX\n", dev_num, dev_addr);
                rc = -ENODEV;
                goto error;
        }

        /* Fill up env_bk */
        p_env = Env + dev_num;
        p_env->devno[READ]  = dev_addr;
        p_env->devno[WRITE] = dev_addr + 1;
        p_env->readBuffers  = num_rbuf;
        p_env->WriteBuffers = num_wbuf;
        p_env->readSize     = CLAW_FRAME_SIZE;
        p_env->write_size    = CLAW_FRAME_SIZE;

        memset(p_env->host_name, 0x20, MAX_NAME_LEN);
        memset(p_env->adapter_name, 0x20, MAX_NAME_LEN);
        memset(p_env->ApiType, 0x20,  MAX_NAME_LEN);
        memcpy(p_env->host_name, hostname, strlen(hostname));
        memcpy(p_env->adapter_name, adaptername, strlen(adaptername));
        memcpy(p_env->ApiType, apitype, strlen(apitype));

        /* Set device name and register */
        dev = claw_devs + dev_num;
        memset(dev->name, 0x00, 8);
        sprintf(dev->name, "claw%ld", dev_num);
#ifdef DEBUGMSG
        printk(KERN_INFO "claw: device name %s\n", dev->name);
#endif 
        if ( (rc = register_netdev(dev)) != 0 ) 
        {
                printk(KERN_WARNING "%s: error %i registering device \"%s\"\n",
                        dev->name,rc, dev->name);
                goto error;
        }
        p_env->dev = dev;


        printk(KERN_INFO "claw%ld:readsize=%d  writesize=%d readbuffer=%d "
	       "writebuffer=%d read=%04x write=%04x\n", 
                dev_num, p_env->readSize, p_env->write_size,p_env->readBuffers,
                p_env->WriteBuffers, p_env->devno[READ], p_env->devno[WRITE]);

        printk(KERN_INFO "claw%ld:host_name:%.8s, adapter_name:%.8s "
	       "ApiType: %.8s\n", 
                dev_num,p_env->host_name, p_env->adapter_name, p_env->ApiType);
                
        ++num_claw_devices;

  error:

#ifdef FUNCTRACE
        printk(KERN_INFO "claw%ld: claw_init_dev() "
	       "> Exit on line %d, rc = %d\n",dev_num, __LINE__, rc);
        printk(KERN_INFO "claw%ld: p_env\n",dev_num);      
        dumpit((char *)p_env, sizeof(CLAW_ENV));
#endif
        return rc;
}
#endif

/*--------------------------------------------------------------------*
*    claw_cleanup                                                     *
*---------------------------------------------------------------------*/

#if !defined(CLAW_CHANDEV) || defined(MODULE)
static void claw_cleanup(void) 
{
        PENV p_env;
        int i;
        int c;
        PPRIV privptr;

#ifdef FUNCTRACE
        printk(KERN_INFO "claw: claw_cleanup() > enter \n");
#endif
#ifdef CLAW_CHANDEV
        chandev_unregister(claw_chandev_probe, 1);
#endif

        p_env=(PENV)&Env;
        for (i=0 ; i< CLAW_MAX_DEV ; i++) {
                while (p_env->dev !=NULL) {
                        privptr = (PPRIV) p_env->dev->priv;
                        for (c = READ; c <= WRITE; ++c)
                        {
                                FREE_IRQ(get_irq_by_devno(p_env->devno[c]),
					 privptr->channel[c].devstat);
                        }
                        unregister_netdev(p_env->dev);
                        probe_error(p_env->dev);
                        p_env->p_priv=NULL;
                        p_env->dev=NULL;
                }
                ++p_env;
        }
        remove_proc_entry("claw", 0);
        printk(KERN_INFO "claw driver unloaded\n");
        
#ifdef FUNCTRACE
        printk(KERN_INFO "claw: claw_cleanup() > exit on line %d\n", __LINE__);
#endif
}
#endif

/*--------------------------------------------------------------------*
*    claw_init                                                        *
*---------------------------------------------------------------------*/

static __init int claw_init(void) 
{
        int i;

        printk(KERN_INFO "Starting CLAW driver "
#ifdef MODULE
                "module "
#else
                "compiled into kernel "
#endif
                " $Revision: 1.17 $ "
#ifdef CLAW_CHANDEV
                      "with"
#else
                      "without" 
#endif
               " chandev support.\n");


#ifdef FUNCTRACE
        printk(KERN_INFO "claw: claw_init() > enter \n");
#endif

        memset(Env, 0, sizeof(Env));
        for (i = 0; i< CLAW_MAX_DEV; i++) 
        {
            Env[i].devno[WRITE] = -ENODEV;
            Env[i].devno[READ]  = -ENODEV;
        }

#ifdef CLAW_CHANDEV

        chandev_register_and_probe(claw_chandev_probe, 
                                (chandev_shutdownfunc)claw_chandev_release,
                                claw_msck_notify, chandev_type_claw);

#else /* CLAW_CHANDEV */
        for (i = 0; i < 256 && claw[i]; i++)
        {
                int rc = claw_init_dev(claw[i]);
                if (rc != 0)
                {
                        printk(KERN_INFO "claw: Please specify "
			       "parameters \n");
                        printk(KERN_INFO "      eg. claw=0:0x0200:5:"
			       "5:LINUX390:RS6K:API\n");
#ifdef FUNCTRACE
                        printk(KERN_INFO "claw: claw_init() "
			       "> exit on line %d, rc = EINVAL\n", __LINE__);
#endif

                        claw_cleanup();
                        return -EINVAL;
                }
        }
#endif /* CLAW_CHANDEV */

        claw_proc_entry = create_proc_entry("claw", 0, 0);
        claw_proc_entry->get_info = claw_read_proc;


#ifdef FUNCTRACE
        printk(KERN_INFO "claw: claw_init() "
	       "> exit on line %d, rc = 0\n", __LINE__);
#endif
        return 0;
}


/*-----------------------------------------------------------*
*      claw_setup                                            *
*------------------------------------------------------------*/

#ifndef MODULE

/* 2.2 - style kernel command line argument parser */
void __init claw_setup(char *str, int *ints)
{
        char *next;
        int i;

        for (i = 0; i < 256; i++)
                if (claw[i] == NULL)
                        break;
        if (i >= 256)
                return;

        next = str;
        while ( (str = next) != NULL )
        {
                next = strchr(str, ',');
                if (next)
                        *next++ = 0;

                claw[i++] = str;
                if (i >= 256)
                        return;
        }
}

/* 2.4 - style kernel command line argument parser */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,18)
int __init claw_call_setup(char *str)
{
        claw_setup(str, NULL);
        return 1;
}
__setup("claw=", claw_call_setup);
__initcall(claw_init);
#endif

#else /* MODULE */

int init_module(void) 
{
        return claw_init();
}

void cleanup_module(void)
{
        claw_cleanup();
}

#endif /* MODULE */




/*--------------------------------------------------------------------*
*    claw_read_proc()                                             *
*                                                                       *
*---------------------------------------------------------------------*/
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
int claw_read_proc(char *buf, char **start, off_t offset, int len, int unused)
#else
int claw_read_proc(char *buf, char **start, off_t offset, int len)
#endif
{
        int i;  
        PENV p_env;
        PPRIV privptr;
        char hostname[9], adaptername[9], devname[9], apitype[9];
        static char *claw_state_string[] = { "STOPPED" , "HALT_IO", 
		"SENSE", "STARTED", "STARTED" };

        len = 0;
        p_env=(PENV) &Env;
        len += sprintf(buf+len, "Linux on zSeries CLAW Driver "
		       "$Revision: 1.17 $\n");                      
        len += sprintf(buf+len, "  Number of CLAW devices:  %d\n",
		       num_claw_devices);

        for (i=0; i < CLAW_MAX_DEV; ++i)
        {
                if (p_env->in_use) {
                        privptr = (PPRIV) p_env->dev->priv;
                        memset(hostname, 0x00, 9);
                        memcpy(hostname, p_env->host_name, 8);
                        memset(adaptername, 0x00, 9);
                        memcpy(adaptername, p_env->adapter_name, 8);
                        memset(apitype, 0x00, 9);
                        memcpy(apitype, p_env->ApiType, 8);
                        memset(devname, 0x00, 9);
                        memcpy(devname, p_env->dev->name, 8);
                        len += sprintf(buf+len, "  Device %s:\n", devname);
                        len += sprintf(buf+len, "      read Channel "
				       "Address:  %#X, IRQ %d  state: %s\n",
				       p_env->devno[READ],
				       privptr->channel[READ].irq,
				       claw_state_string[privptr->
				       channel[READ].claw_state]);
                        len += sprintf(buf+len, "      Write Channel "
				       "Address: %#X, IRQ %d  state: %s\n", 
				       p_env->devno[WRITE],
				       privptr->channel[WRITE].irq,
				       claw_state_string[privptr->
				       channel[WRITE].claw_state]);
                        len += sprintf(buf+len, "      Hostname:    %s\n",
				       hostname);
                        len += sprintf(buf+len, "      Adaptername: %s\n",
				       adaptername);
                        len += sprintf(buf+len, "      Apitype:     %s\n",
				       apitype);
                        len += sprintf(buf+len, "      read Buffers:  %d "
				       "- %d byte buffers\n",
				       p_env->readBuffers,
				       p_env->readSize);
                        len += sprintf(buf+len, "      Write Buffers: %d "
				       "- %d byte buffers   free count: %d\n",
				       p_env->WriteBuffers,
				       p_env->write_size,
				       privptr->write_free_count);
                        len += sprintf(buf+len, "      Interface MTU: %d\n",
				       p_env->dev->mtu);
                }
                ++p_env;
        }

        return len;
}
